#!/usr/bin/env python
"""Manage AWS IAM user login profile.

Usage:
  awsloginprofile USER [--config FILE]
                       [--master-account-id ID]
                       [--auth-account-id ID]
                       [--org-access-role ROLE]
                       [--report]
                       [--new | --reset | --reenable]
                       [--no-email]
                       [--disable]
                       [--disable-expired]
                       [--opt-ttl HOURS]
                       [--password PASSWORD]
                       [-q] [-d|-dd]
  awsloginprofile (--help|--version)

Options:
  USER                      Name of IAM user.
  -h, --help                Show this help message and exit.
  -V, --version             Display version info and exit.
  --config FILE             AWS Org config file in yaml format.
  --master-account-id ID    AWS account Id of the Org master account.
  --auth-account-id ID      AWS account Id of the authentication account.
  --org-access-role ROLE    IAM role for traversing accounts in the Org.
  --report                  Print user login profile report.  this is the default
  --new                     Create new login profile.
  --reset                   Reset password for existing login profile.
  --no-email                Do not email user when (re)setting login profile.
  --disable                 Delete existing login profile, disable access keys.
  --disable-expired         Delete profile if one-time-password exceeds --opt-ttl.
  --reenable                Recreate login profile, reactivate access keys.
  --opt-ttl HOURS           One-time-password time to live in hours [default: 24].
  --password PASSWORD       Supply password, do not require user to reset.
  -q, --quiet               Repress log output.
  -d, --debug               Increase log level to 'DEBUG'.
  -dd                       Include botocore and boto3 logs in log stream.

"""


import os
import sys
import yaml
import logging
from string import Template
import datetime
import smtplib
from email.message import EmailMessage


import boto3
from botocore.exceptions import ClientError
from docopt import docopt
from passwordgenerator import pwgenerator


import awsorgs
from awsorgs.utils import *
from awsorgs.spec import *
from awsorgs.reports import *


# Relative path within awsorgs project to template file used by prep_email()
EMAIL_TEMPLATE = 'data/email_template'


def utcnow():
    return datetime.datetime.now(datetime.timezone.utc)


def get_user_name():
    """
    Returns the IAM user_name of the calling identidy (i.e. you)
    """
    sts = boto3.client('sts')
    return sts.get_caller_identity()['Arn'].split('/')[-1]


def list_delegations(log, user, deployed_accounts):
    """
    Return list of assume role resource arns for user.  Obtain these by
    parsing in-line group policies for each group the user is a member of.

    the policies we care about match one of two patterns:
        AllowAssumeRole-<rolename>
        DenyAssumeRole-<rolename>

    we assemble a list of allowed role arns, then we remove any of the
    denied arns.

    if the account_id field of a AllowAssumeRole-<rolename> matches the
    glob ('*') char, we generate the list of allowed role arns - one for
    every 'ACTIVE' account.
    """
    groups = list(user.groups.all())
    role_arns = []
    for group in user.groups.all():

        allow_policy = next((
            p for p in list(group.policies.all())
            if p.policy_name.startswith('AllowAssumeRole')
        ), None)
        if allow_policy is not None:
            allow_arns = allow_policy.policy_document['Statement'][0]['Resource']

            if isinstance(allow_arns, str) and '*' in allow_arns:
                head, sep, tail = allow_arns.partition('*')
                allow_arns = []
                for account in deployed_accounts:
                    if account['Status'] == 'ACTIVE':
                        allow_arns.append(head + account['Id'] + tail)

            deny_policy = next((
                p for p in list(group.policies.all())
                if p.name.startswith('DenyAssumeRole')
            ), None)
            if deny_policy is not None:
                deny_arns = deny_policy.policy_document['Statement'][0]['Resource']
                deny_account_ids = [arn.split(':')[4] for arn in deny_arns]
                for id in deny_account_ids:
                    for arn in allow_arns:
                        if id in arn:
                            allow_arns.remove(arn)

            role_arns += allow_arns

    return role_arns


def format_delegation_table(delegation_arns, deployed_accounts):
    """Generate formatted list of delegation attributes as printable string"""
    template = "  {}    {}{}{}\n"
    delegation_string = template.format('Account Id  ', 'Alias', ' '*19, 'Role')
    for assume_role_arn in delegation_arns:
        account_id = assume_role_arn.split(':')[4]
        alias = lookup(deployed_accounts, 'Id', account_id, 'Alias')
        if alias is None:
            alias = str()
        spacer = (24 - len(alias)) * ' '
        delegation_string += template.format(account_id, alias, spacer,
                assume_role_arn.partition('role/')[2])
    return delegation_string


def user_report(log, deployed_accounts, user, login_profile):
    """Generate report of IAM user's login profile, password usage, and
    assume_role delegations for any groups user is member of.
    """
    spacer = '{:<24}{}'
    log.info('\n')
    log.info(spacer.format('User:', user.name))
    log.info(spacer.format('Arn:', user.arn))
    log.info(spacer.format('User Id:', user.user_id))
    log.info(spacer.format('User created:', user.create_date))
    if login_profile:
        log.info(spacer.format('Login profile created:', login_profile.create_date))
        log.info(spacer.format('Passwd reset required:', login_profile.password_reset_required))
        if login_profile.password_reset_required:
            log.info(spacer.format(
                'One-time-passwd age:',
                utcnow() - login_profile.create_date,
            ))
        else:
            log.info(spacer.format('Password last used:', user.password_last_used))
    else:
        log.info(spacer.format('User login profile:', login_profile))
    assume_role_arns = list_delegations(log, user, deployed_accounts)
    if assume_role_arns:
        log.info('Delegations:\n{}'.format(
            format_delegation_table(assume_role_arns, deployed_accounts)
        ))


def validate_user(user_name, credentials=None):
    """Return a valid IAM User object"""
    if credentials:
        iam = boto3.resource('iam', **credentials)
    else:
        iam = boto3.resource('iam')
    user = iam.User(user_name)
    try:
        user.load()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            return    
    return user


def validate_login_profile(user):
    """Return a valid IAM LoginProfile object"""
    login_profile = user.LoginProfile()
    try:
        login_profile.load()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            return
    return login_profile


def munge_passwd(passwd=None):
    """Return new 'passwd' string and boolean 'require_reset'.
    If passwd provided, set 'require_reset' to False.
    """
    if passwd:
        require_reset = False
    else:
        passwd = pwgenerator.generate()
        require_reset = True
    return passwd, require_reset


def create_profile(log, user, passwd, require_reset):
    log.debug('creating login profile for user %s' % user.name)
    return user.create_login_profile(
        Password=passwd,
        PasswordResetRequired=require_reset,
    )


def reset_profile(log, user, login_profile, passwd, require_reset):
    """Reset IAM user passwd by deleting and recreating login profile.
    This ensures the password creation date gets reset when updating a password.
    """
    if login_profile:
        log.debug('resetting login profile for user %s' % user.name)
        login_profile.delete()
        return login_profile.create(
            Password=passwd,
            PasswordResetRequired=require_reset
        )
    else:
        log.error("user '%s' has no login profile" % user.name)
        sys.exit(1)

def delete_profile(log, user, login_profile):
    if login_profile:
        log.info('deleting login profile for user %s' % user.name)
        login_profile.delete()
    else:
        log.warn("user '%s' has no login profile" % user.name)


def set_access_key_status(log, user, enable=True):
    """Enable or disable an IAM user's access keys"""
    for key in user.access_keys.all():
        if enable and key.status == 'Inactive':
            log.info('enabling access key %s for user %s' %
                    (key.access_key_id, user.name))
            key.activate()
        elif not enable and key.status == 'Active':
            log.info('disabling access key %s for user %s' %
                    (key.access_key_id, user.name))
            key.deactivate()


def onetime_passwd_expired(log, user, login_profile, hours):
    """Test if initial one-time-only password is expired"""
    if login_profile and login_profile.password_reset_required:
        log.debug('now: %s' % utcnow().isoformat())
        log.debug('ttl: %s' % datetime.timedelta(hours=hours))
        log.debug('delta: %s' % (utcnow() - login_profile.create_date))
        return (utcnow() - login_profile.create_date) > datetime.timedelta(hours=hours)
    return False


def prep_email(log, aliases, deployed_accounts, user, passwd):
    """Generate email body from template"""
    log.debug("loading file: '%s'" % EMAIL_TEMPLATE)
    trusted_id=boto3.client('sts').get_caller_identity()['Account']
    if aliases:
        trusted_account = aliases[trusted_id]
    else:
        trusted_account = trusted_id
    assume_role_arns = list_delegations(log, user, deployed_accounts)
    log.debug('assume_role_arns: %s' % assume_role_arns)
    template = os.path.abspath(pkg_resources.resource_filename(__name__, EMAIL_TEMPLATE))
    mapping = dict(
        user_name=user.name,
        onetimepw=passwd,
        trusted_account=trusted_account,
        delegations=format_delegation_table(assume_role_arns, deployed_accounts),
    )
    with open(template) as tpl:
        return Template(tpl.read()).substitute(mapping)


def build_email_message(user, message_body, spec):
    org_admin_team = lookup(spec['teams'], 'Name', spec['org_admin_team'])
    msg = EmailMessage()
    msg.set_content(message_body)
    msg['Subject'] = 'login profile'
    msg['To'] = lookup(spec['users'], 'Name', user.name, 'Email')
    msg['From'] = ', '.join(org_admin_team['TechnicalContacts'])
    msg['Cc'] = ', '.join(org_admin_team['BusinessContacts'])
    return msg

def send_email(msg, smtp_server):
    s = smtplib.SMTP(smtp_server)
    s.send_message(msg)
    s.quit()


def handle_email(log, args, spec, aliases, deployed_accounts, user, passwd):
    message_body = prep_email(log, aliases, deployed_accounts, user, passwd)
    if args['--no-email']:
        print(message_body)
    else:
        msg = build_email_message(user, message_body, spec)
        send_email(msg, spec['default_smtp_server'])


def main():
    args = docopt(__doc__, version=awsorgs.__version__)
    # HACK ALERT!
    # set '--exec' and 'report' args to make get_logger() happy
    args['--exec'] = True
    if not (args['--new']
            or args['--reset']
            or args['--disable']
            or args['--disable-expired']
            or args['--reenable']):
        args['report'] = True
    else:
        args['report'] = False
    log = get_logger(args)
    log.debug("%s: args:\n%s" % (__name__, args))
    args = load_config(log, args)
    spec = validate_spec(log, args)

    user = validate_user(args['USER'])
    if not user:
        log.critical('no such user: %s' % args['USER'])
        sys.exit(1)
    login_profile = validate_login_profile(user)
    passwd, require_reset = munge_passwd(args['--password'])
    org_credentials = get_assume_role_credentials(
            args['--master-account-id'],
            args['--org-access-role'])
    if isinstance(org_credentials, RuntimeError):
        log.critical(org_credentials)
        sys.exit(1)
    org_client = boto3.client('organizations', **org_credentials)
    deployed_accounts = scan_deployed_accounts(log, org_client)
    aliases = get_account_aliases(log, deployed_accounts, args['--org-access-role'])
    deployed_accounts = merge_aliases(log, deployed_accounts, aliases)
    log.debug(aliases)

    if args['--new']:
        if not login_profile:
            login_profile = create_profile(log, user, passwd, require_reset)
            handle_email(log, args, spec, aliases, deployed_accounts, user, passwd)
        else:
            log.warn("login profile for user '%s' already exists" % user.name)
            user_report(log, deployed_accounts, user, login_profile)

    elif args['--reset']:
        login_profile = reset_profile(log, user, login_profile, passwd, require_reset)
        handle_email(log, args, spec, aliases, deployed_accounts, user, passwd)

    elif args['--disable']:
        delete_profile(log, user, login_profile)
        set_access_key_status(log, user, False)

    elif args['--disable-expired']:
        if onetime_passwd_expired(log, user, login_profile, int(args['--opt-ttl'])):
            delete_profile(log, user, login_profile)

    elif args['--reenable']:
        if not login_profile:
            login_profile = create_profile(log, user, passwd, require_reset)
            handle_email(log, args, spec, aliases, deployed_accounts, user, passwd)
        else:
            log.warn("login profile for user '%s' already exists" % user.name)
        set_access_key_status(log, user, True)

    else:
        user_report(log, deployed_accounts, user, login_profile)


if __name__ == "__main__":
    main()



"""Utility functions used by the various awsorgs modules"""

import os
import sys
import re
import pkg_resources
import difflib
import threading
try:
    import queue
except ImportError:
    import Queue as queue

import boto3
from botocore.exceptions import ClientError
import yaml
import logging


S3_BUCKET_PREFIX = 'awsorgs'
S3_OBJECT_KEY = 'deployed_accounts.yaml'


def get_s3_bucket_name(prefix=S3_BUCKET_PREFIX):
    """
    Generate an s3 bucket name based on a name prefix and the aws account ig
    """
    sts_client = boto3.client('sts')
    account_id = sts_client.get_caller_identity()['Account']
    return '-'.join([prefix, account_id])


def lookup(dlist, lkey, lvalue, rkey=None):
    """
    Use a known key:value pair to lookup a dictionary in a list of
    dictionaries.  Return the dictonary or None.  If rkey is provided,
    return the value referenced by rkey or None.  If more than one
    dict matches, raise an error.
    args:
        dlist:   lookup table -  a list of dictionaries
        lkey:    name of key to use as lookup criteria
        lvalue:  value to use as lookup criteria
        rkey:    (optional) name of key referencing a value to return
    """
    items = [d for d in dlist
             if lkey in d
             and d[lkey] == lvalue]
    if not items:
        return None
    if len(items) > 1:
        raise RuntimeError(
            "Data Error: lkey: {}, lvalue: {} - lookup matches multiple items in dlist".format(lkey, lvalue)
        )
    if rkey:
        if rkey in items[0]:
            return items[0][rkey]
        return None
    return items[0] 

def search_spec(spec, search_key, recurse_key):
    """
    Recursively scans spec structure and returns a list of values
    keyed with 'search_key' or and empty list.  Assumes values
    are either list or str.
    """
    value = []
    if search_key in spec and spec[search_key]:
        if isinstance(spec[search_key], str):
            value.append(spec[search_key])
        else:
            value += spec[search_key]
    if recurse_key in spec and spec[recurse_key]:
        for child_spec in spec[recurse_key]:
            value += search_spec(child_spec, search_key, recurse_key)
    return sorted(value)


def ensure_absent(spec):
    """
    test if an 'Ensure' key is set to absent in dictionary 'spec'
    """
    if 'Ensure' in spec and spec['Ensure'] == 'absent': return True
    return False


def munge_path(default_path, spec):
    """
    Return formated 'Path' attribute for use in iam client calls. 
    Unless specified path is fully qualified (i.e. starts with '/'),
    prepend the 'default_path'.
    """
    if 'Path' in spec and spec['Path']:
        if spec['Path'][0] == '/':
            if spec['Path'][-1] != '/':
                return spec['Path'] + '/'
            return spec['Path']
        return "/%s/%s/" % (default_path, spec['Path'])
    return "/%s/" % default_path


def get_logger(args):
    """
    Setup logging.basicConfig from args.
    Return logging.Logger object.
    """
    # log level
    log_level = logging.INFO
    if args['--debug']:
        log_level = logging.DEBUG
    if args['--quiet']:
        log_level = logging.CRITICAL
    # log format
    log_format = '%(name)s: %(levelname)-9s%(message)s'
    if args['report']:
        log_format = '%(message)s'
    if args['--debug'] == 1:
        log_format = '%(name)s: %(levelname)-9s%(funcName)s():  %(message)s'
    if (not args['--exec'] and not args['report']):
        log_format = '[dryrun] %s' % log_format
    if not args['--debug'] == 2:
        logging.getLogger('botocore').propagate = False
        logging.getLogger('boto3').propagate = False
    logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    log = logging.getLogger(__name__)
    return log


def valid_account_id(log, account_id):
    """Validate account Id is a 12 digit string"""
    if not isinstance(account_id, str):
        log.error("supplied account id {} is not a string".format(account_id))
        return False
    id_re = re.compile(r'^\d{12}$')
    if not id_re.match(account_id):
        log.error("supplied account id '{}' must be a 12 digit number".format(account_id))
        return False
    return True


def get_root_id(org_client):
    """
    Query deployed AWS Organization for its Root ID.
    """
    roots = org_client.list_roots()['Roots']
    if len(roots) >1:
        raise RuntimeError("org_client.list_roots returned multiple roots.")
    return roots[0]['Id']


def validate_master_id(org_client, spec):
    """
    Don't mangle the wrong org by accident
    """
    master_account_id = org_client.describe_organization(
      )['Organization']['MasterAccountId']
    if master_account_id != spec['master_account_id']:
        errmsg = ("The Organization Master Account Id '%s' does not match the "
                "'master_account_id' set in the spec-file" % master_account_id)
        raise RuntimeError(errmsg)
    return


def queue_threads(log, sequence, func, f_args=(), thread_count=20):
    """generalized abstraction for running queued tasks in a thread pool"""

    def worker(*args):
        log.debug('%s: q.empty: %s' % (threading.current_thread().name, q.empty()))
        while not q.empty():
            log.debug('%s: task: %s' % (threading.current_thread().name, func))
            item = q.get()
            log.debug('%s: processing item: %s' % (threading.current_thread().name, item))
            func(item, *args)
            q.task_done()

    q = queue.Queue()
    for item in sequence:
        log.debug('queuing item: %s' % item)
        q.put(item)
    log.debug('queue length: %s' % q.qsize())
    for i in range(thread_count):
        t = threading.Thread(target=worker, args=f_args)
        t.setDaemon(True)
        t.start()
    q.join()


def get_assume_role_credentials(account_id, role_name, region_name=None):
    """
    Get temporary sts assume_role credentials for account.
    """
    role_arn = "arn:aws:iam::%s:role/%s" % (account_id, role_name)
    role_session_name = account_id + '-' + role_name.split('/')[-1]
    sts_client = boto3.client('sts')

    if account_id == sts_client.get_caller_identity()['Account']:
        return dict(
                aws_access_key_id=None,
                aws_secret_access_key=None,
                aws_session_token=None,
                region_name=None)
    else:
        try:
            credentials = sts_client.assume_role(
                    RoleArn=role_arn,
                    RoleSessionName=role_session_name
                    )['Credentials']
        except ClientError as e:
            if e.response['Error']['Code'] == 'AccessDenied':
                errmsg = ('cannot assume role %s in account %s' %
                        (role_name, account_id))
                return RuntimeError(errmsg)
        return dict(
                aws_access_key_id=credentials['AccessKeyId'],
                aws_secret_access_key=credentials['SecretAccessKey'],
                aws_session_token=credentials['SessionToken'],
                region_name=region_name)


def scan_deployed_accounts(log, org_client):
    """
    Query AWS Organization for deployed accounts.
    Returns a list of dictionary.
    """
    log.debug('running')
    accounts = org_client.list_accounts()
    deployed_accounts = accounts['Accounts']
    while 'NextToken' in accounts and accounts['NextToken']:
        accounts = org_client.list_accounts(NextToken=accounts['NextToken'])
        deployed_accounts += accounts['Accounts']
    # only return accounts that have an 'Name' key
    return [d for d in deployed_accounts if 'Name' in d ]


def scan_created_accounts(log, org_client):
    """
    Query AWS Organization for accounts with creation status of 'SUCCEEDED'.
    Returns a list of dictionary.
    """
    log.debug('running')
    status = org_client.list_create_account_status(States=['SUCCEEDED'])
    created_accounts = status['CreateAccountStatuses']
    while 'NextToken' in status and status['NextToken']:
        status = org_client.list_create_account_status(
                States=['SUCCEEDED'],
                NextToken=status['NextToken'])
        created_accounts += status['CreateAccountStatuses']
    return created_accounts
        

def get_account_aliases(log, deployed_accounts, role):
    """
    Return dict of {Id:Alias} for all deployed accounts.

    role::  name of IAM role to assume to query all deployed accounts.
    """
    # worker function for threading
    def get_account_alias(account, log, role, aliases):
        if account['Status'] == 'ACTIVE':
            credentials = get_assume_role_credentials(account['Id'], role)
            if isinstance(credentials, RuntimeError):
                log.error(credentials)
                return
            iam_client = boto3.client('iam', **credentials)
            response = iam_client.list_account_aliases()['AccountAliases']
            if response:
                aliases[account['Id']] = response[0]
    # call workers
    aliases = {}
    queue_threads(log, deployed_accounts, get_account_alias,
            f_args=(log, role, aliases), thread_count=10)
    log.debug(yamlfmt(aliases))
    return aliases


def merge_aliases(log, deployed_accounts, aliases):
    """
    Merge account aliases into deployed_accounts lookup table.
    """
    for account in deployed_accounts:
        account['Alias'] = aliases.get(account['Id'], '')
        log.debug(account)
    return deployed_accounts


def string_differ(string1, string2):
    """Returns the diff of 2 strings"""
    diff = difflib.ndiff(
        string1.splitlines(keepends=True),
        string2.splitlines(keepends=True),
    )
    return ''.join(list(diff))


def yamlfmt(dict_obj):
    """Convert a dictionary object into a yaml formated string"""
    return yaml.dump(dict_obj, default_flow_style=False)


def overbar(string):
    """
    Returns string preceeded by an overbar of the same length:
    >>> print(overbar('blee'))
    ____
    blee
    """
    return "%s\n%s" % ('_' * len(string), string)


def report_maker(log, accounts, role, query_func, report_header=None, **qf_args):
    """
    Generate a report by running a arbitrary query function in each account.
    The query function must return a list of strings.
    """
    # Thread worker function to gather report for each account
    def make_account_report(account, report, role):
        messages = []
        messages.append(overbar("Account:    %s" % account['Name']))
        credentials = get_assume_role_credentials(account['Id'], role)
        if isinstance(credentials, RuntimeError):
            messages.append(credentials)
        else:
            messages += query_func(credentials, **qf_args)
        report[account['Name']] = messages
    # gather report data from accounts
    report = {}
    queue_threads(
            log, accounts,
            make_account_report,
            f_args=(report, role),
            thread_count=10)
    # process the reports
    if report_header:
        log.info("\n\n%s" % overbar(report_header))
    for account, messages in sorted(report.items()):
        for msg in messages:
            log.info(msg)


def get_iam_objects(iam_client_function, object_key, f_args=dict()):
    """
    users = get_iam_objects(iam_client.list_users, 'Users')
    """
    iam_objects = []
    response = iam_client_function(**f_args)
    iam_objects += response[object_key]
    if 'IsTruncated' in response:
        while response['IsTruncated']:
            response = iam_client_function(Marker=response['Marker'],**f_args)
            iam_objects += response[object_key]
    return iam_objects

    




"""
Report maker utility and query functions

Todo:
    allow reporting on single account or short list of accounts
    substitute account alias for account id in reports
    
"""

import io
import csv
from awsorgs.utils import *


# Report_maker utilities

def overbar(string):
    """
    Returns string preceeded by an overbar of the same length:
    >>> print(overbar('blee'))
    ____
    blee
    """
    return "%s\n%s" % ('_' * len(string), string)


def report_maker(log, accounts, role, query_func, report_header=None, **qf_args):
    """
    Generate a report by running a arbitrary query function in each account.
    The query function must return a list of strings.
    """
    # Thread worker function to gather report for each account
    def make_account_report(account, report, role):
        messages = []
        messages.append(overbar("Account:    %s" % account['Name']))
        credentials = get_assume_role_credentials(account['Id'], role)
        if isinstance(credentials, RuntimeError):
            messages.append(credentials)
        else:
            messages += query_func(credentials, **qf_args)
        report[account['Name']] = messages
    # gather report data from accounts
    report = {}
    queue_threads(
            log, accounts,
            make_account_report,
            f_args=(report, role),
            thread_count=10)
    # process the reports
    if report_header:
        log.info("\n\n%s" % overbar(report_header))
    for account, messages in sorted(report.items()):
        for msg in messages:
            log.info(msg)

    
# report_maker query functions

def user_group_report(credentials, verbose=False):
    """
    A report_maker query function.
    Reports IAM users and Groups in an account.

    ISSUE: report access keys, ssh keys, mfa devices, http users

    """
    messages = []
    iam_client = boto3.client('iam', **credentials)

    user_info = []
    users = get_iam_objects(iam_client.list_users, 'Users')
    for u in users:
        if verbose:
            user_info.append(u)
        else:
            user_info.append(u['Arn'])
    if user_info:
        messages.append(yamlfmt(dict(Users=user_info)))

    group_info = []
    groups = get_iam_objects(iam_client.list_groups, 'Groups')
    for g in groups:
        if verbose:
            group_info.append(g)
        else:
            group_info.append(g['Arn'])
    if group_info:
        messages.append(yamlfmt(dict(Groups=group_info)))
    #if groups:
    #    messages.append("Groups:")
    #    if verbose:
    #        messages.append(yamlfmt(groups))
    #    else:
    #        messages += ["  %s" % group['Arn'] for group in groups]
    return messages


def credentials_report(credentials):
    """
    A report_maker query function.
    IAM Credential report in an account

    ISSUES:
      Clean up exception handling:
      botocore.errorfactory.CredentialReportNotPresentException: An error occurred (ReportNotPresent) when calling the GetCredentialReport operation: Unknown
    """

    messages = []
    iam_client = boto3.client('iam', **credentials)
    try:
        response = iam_client.get_credential_report()
    except Exception as e:
        response = iam_client.generate_credential_report()
        messages.append(yamlfmt(response))
        return messages

    report_file_object = io.StringIO(response['Content'].decode())
    reader = csv.DictReader(report_file_object)
    user_info = []
    for row in reader:
        user = dict()
        for key in reader.fieldnames:
            user['UserName'] = row['user']
            user['Arn'] = row['arn']
            if (key not in ['user', 'arn'] and 
                    row[key] not in ['N/A', 'not_supported', 'no_information', 'false']):
                user[key] = row[key]
        user_info.append(user)

    if user_info:
        messages.append(yamlfmt(dict(Users=user_info)))
    return messages


def role_report(credentials, verbose=False):
    """
    A report_maker query function.
    Reports IAM custom policies and roles in an account.
    """
    messages = []
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)

    policy_info = []
    custom_policies = get_iam_objects(iam_client.list_policies, 'Policies', 
            dict(Scope='Local'))
    for p in custom_policies:
        if verbose:
            policy_version_id = iam_resource.Policy(p['Arn']).default_version_id
            policy_info.append(dict(
                Arn=p['Arn'],
                Statement=iam_resource.PolicyVersion(
                    p['Arn'], 
                    policy_version_id
                ).document['Statement'],
            ))
        else:
            policy_info.append(p['Arn'])
    if policy_info:
        messages.append(yamlfmt(dict(CustomPolicies=policy_info)))

    role_info = []
    roles = get_iam_objects(iam_client.list_roles, 'Roles')
    for r in roles:
        role = iam_resource.Role(r['RoleName'])
        if verbose:
            role_info.append(dict(
                Arn=role.arn,
                Statement=role.assume_role_policy_document['Statement'],
                AttachedPolicies=[p.policy_name for p in list(role.attached_policies.all())],
            ))
        else:
            role_info.append(role.arn)
    if role_info:
        messages.append(yamlfmt(dict(Roles=role_info)))
    return messages


def account_authorization_report(credentials, verbose=False):
    """
    A report_maker query function.
    IAM Account Authorization Reporting

    """
    messages = []
    iam_client = boto3.client('iam', **credentials)

    user_info = []
    users = get_iam_objects(
            iam_client.get_account_authorization_details, 
            'UserDetailList',
            dict(Filter=['User']))
    for u in users:
        if verbose:
            user_info.append(u)
        else:
            user_info.append(u['Arn'])
    if user_info:
        messages.append(yamlfmt(dict(Users=user_info)))

    group_info = []
    groups = get_iam_objects(
            iam_client.get_account_authorization_details, 
            'GroupDetailList',
            dict(Filter=['Group']))
    for u in groups:
        if verbose:
            group_info.append(u)
        else:
            group_info.append(u['Arn'])
    if group_info:
        messages.append(yamlfmt(dict(Groups=group_info)))

    role_info = []
    roles = get_iam_objects(
            iam_client.get_account_authorization_details, 
            'RoleDetailList',
            dict(Filter=['Role']))
    for u in roles:
        if verbose:
            role_info.append(u)
        else:
            role_info.append(u['Arn'])
    if role_info:
        messages.append(yamlfmt(dict(Roles=role_info)))

    policy_info = []
    policies = get_iam_objects(
            iam_client.get_account_authorization_details, 
            'Policies',
            dict(Filter=['LocalManagedPolicy']))
    for u in policies:
        if verbose:
            policy_info.append(u)
        else:
            policy_info.append(u['Arn'])
    if policy_info:
        messages.append(yamlfmt(dict(CustomPolicies=policy_info)))


    return messages




# Obsolete resource display functions. For reference only
#
#display_provisioned_users(log, args, deployed, auth_spec, credentials)
#display_provisioned_groups(log, args, deployed, credentials)
#display_roles_in_accounts(log, args, deployed, auth_spec)


def display_provisioned_users(log, args, deployed, auth_spec, credentials):
    """
    Print report of currently deployed IAM users in Auth account.
    """
    header = "Provisioned IAM Users in Auth Account:"
    overbar = '_' * len(header)
    log.info("\n%s\n%s\n" % (overbar, header))
    if args['--full']:
        aliases = get_account_aliases(log, deployed['accounts'],
                args['--org-access-role'])
    for name in sorted([u['UserName'] for u in deployed['users']]):
        arn = lookup(deployed['users'], 'UserName', name, 'Arn')
        if args['--full']:
            user = validate_user(name, credentials)
            if user:
                login_profile = validate_login_profile(user)
                user_report(log, aliases, user, login_profile)
        else:
            spacer = ' ' * (12 - len(name))
            log.info("%s%s\t%s" % (name, spacer, arn))


def display_provisioned_groups(log, args, deployed, credentials):
    """
    Print report of currently deployed IAM groups in Auth account.
    List group memebers, attached policies and delegation assume role
    profiles.
    """
    # Thread worker function to assemble lines of a group report
    def display_group(group_name, report, iam_resource):
        log.debug('group_name: %s' % group_name)
        messages = []
        group = iam_resource.Group(group_name)
        members = list(group.users.all())
        attached_policies = list(group.attached_policies.all())
        assume_role_resources = [p.policy_document['Statement'][0]['Resource']
                for p in list(group.policies.all()) if
                p.policy_document['Statement'][0]['Action'] == 'sts:AssumeRole']
        overbar = '_' * (8 + len(group_name))
        messages.append('\n%s' % overbar)
        messages.append("%s\t%s" % ('Name:', group_name))
        messages.append("%s\t%s" % ('Arn:', group.arn))
        if members:
            messages.append("Members:")
            messages.append("\n".join(["  %s" % u.name for u in members]))
        if attached_policies:
            messages.append("Policies:")
            messages.append("\n".join(["  %s" % p.arn for p in attached_policies]))
        if assume_role_resources:
            messages.append("Assume role profiles:")
            messages.append("  Account\tRole ARN")
            profiles = {}
            for role_arn in assume_role_resources:
                account_name = lookup(deployed['accounts'], 'Id',
                        role_arn.split(':')[4], 'Name')
                if account_name:
                    profiles[account_name] = role_arn
            for account_name in sorted(profiles.keys()):
                messages.append("  %s:\t%s" % (account_name, profiles[account_name]))
        report[group_name] = messages

    group_names = sorted([g['GroupName'] for g in deployed['groups']])
    log.debug('group_names: %s' % group_names)
    header = "Provisioned IAM Groups in Auth Account:"
    overbar = '_' * len(header)
    log.info("\n\n%s\n%s" % (overbar, header))

    # log report
    if args['--full']:
        # gather report data from groups
        report = {}
        iam_resource = boto3.resource('iam', **credentials)
        queue_threads(log, group_names, display_group, f_args=(report, iam_resource),
                thread_count=10)
        for group_name, messages in sorted(report.items()):
            for msg in messages:
                log.info(msg)
    else:
        # just print the arns
        log.info('')
        for name in group_names:
            arn = lookup(deployed['groups'], 'GroupName', name, 'Arn')
            spacer = ' ' * (12 - len(name))
            log.info("%s%s\t%s" % (name, spacer, arn))


def display_roles_in_accounts(log, args, deployed, auth_spec):
    """
    Print report of currently deployed delegation roles in each account
    in the Organization.
    We only care about AWS principals, not Service principals.
    """
    # Thread worker function to gather report for each account
    def display_role(account, report, auth_spec):
        messages = []
        overbar = '_' * (16 + len(account['Name']))
        messages.append('\n%s' % overbar)
        messages.append("Account:\t%s" % account['Name'])
        credentials = get_assume_role_credentials(
                account['Id'],
                args['--org-access-role'])
        if isinstance(credentials, RuntimeError):
            messages.append(credentials)
        else:
            iam_client = boto3.client('iam', **credentials)
            iam_resource = boto3.resource('iam', **credentials)
            roles = [r for r in iam_client.list_roles()['Roles']]
            custom_policies = iam_client.list_policies(Scope='Local')['Policies']
            if custom_policies:
                messages.append("Custom Policies:")
                for policy in custom_policies:
                    messages.append("  %s" % policy['Arn'])
            messages.append("Roles:")
            for r in roles:
                role = iam_resource.Role(r['RoleName'])
                if not args['--full']:
                    messages.append("  %s" % role.arn)
                else:
                    principal = role.assume_role_policy_document['Statement'][0]['Principal']
                    if 'AWS' in principal:
                        messages.append("  %s" % role.name)
                        messages.append("    Arn:\t%s" % role.arn)
                        messages.append("    Principal:\t%s" % principal['AWS'])
                        attached = [p.policy_name for p
                                in list(role.attached_policies.all())]
                        if attached:
                            messages.append("    Attached Policies:")
                            for policy in attached:
                                messages.append("      %s" % policy)
        report[account['Name']] = messages

    # gather report data from accounts
    report = {}
    queue_threads(log, deployed['accounts'], display_role, f_args=(report, auth_spec),
            thread_count=10)
    # process the reports
    header = "Provisioned IAM Roles in all Org Accounts:"
    overbar = '_' * len(header)
    log.info("\n\n%s\n%s" % (overbar, header))
    for account, messages in sorted(report.items()):
        for msg in messages:
            log.info(msg)



#!/usr/bin/env python
"""Generate default org access role in an invited account.
Run this with IAM credentials for invited account.

Creates role 'OrganizationAccountAccessRole' allowing users in
Org Master account 'AdministratorAccess' in invited account.

Usage:
  awsorgs-accessrole --master_id ID [--exec]
  awsorgs-accessrole --help
  awsorgs-accessrole --version

Options:
  -m, --master_id ID    Master Account ID
  -h, --help            Show this help message and exit.
  -V, --version         Display version info and exit.
"""

import os
import sys
import yaml
import json

import boto3
from botocore.exceptions import ClientError
from docopt import docopt

import awsorgs
from awsorgs.utils import *

ROLENAME = 'OrganizationAccountAccessRole'
DESCRIPTION = 'Organization Access Role'
POLICYNAME = 'AdministratorAccess'

def main():
    args = docopt(__doc__, version=awsorgs.__version__)
    iam_client = boto3.client('iam')
    # assemble assume-role policy statement
    principal = "arn:aws:iam::%s:root" % args['--master_id']
    statement = dict(
            Effect='Allow',
            Principal=dict(AWS=principal),
            Action='sts:AssumeRole')
    policy_doc = json.dumps(dict(
            Version='2012-10-17', Statement=[statement]))
    # create role
    print("Creating role %s" % ROLENAME)
    if args['--exec']:
        iam_client.create_role(
                Description=DESCRIPTION,
                RoleName=ROLENAME,
                AssumeRolePolicyDocument=policy_doc)
    # attach policy to new role
    iam_resource = boto3.resource('iam')
    aws_policies = iam_client.list_policies(Scope='AWS',
            MaxItems=500)['Policies']
    policy_arn = lookup(aws_policies, 'PolicyName', POLICYNAME, 'Arn')
    role = iam_resource.Role(ROLENAME)
    try:
        role.load()
    except:
        pass
    else:
        print("Attaching policy %s to %s" % (POLICYNAME, ROLENAME))
        if args['--exec'] and policy_arn:
            role.attach_policy(PolicyArn=policy_arn)


if __name__ == "__main__":
    main()



"""
Spec validator schema data

ISSUES:
    place regex rule on email addresses, domain name
"""
import yaml

from cerberus import Validator, schema_registry
from awsorgs.utils import yamlfmt


# Schema for validating spec files.  Since spec is accumulated from multiple
# files, we do not place a 'require' rule on first level keys.  Individual spec
# files have only a subset of these.
#
SPEC_FILE_SCHEMA = """
master_account_id:
  type: string
auth_account_id:
  type: string
default_domain:
  type: string
default_sc_policy:
  type: string
default_ou:
  type: string
default_path:
  type: string
default_smtp_server:
  type: string
org_admin_team:
  type: string
organizational_units:
  required: False
  type: list
  schema:
    type: dict
    schema: organizational_unit
sc_policies:
  required: False
  type: list
  schema:
    type: dict
    schema: sc_policy
teams:
  required: False
  type: list
  schema:
    type: dict
    schema: team
accounts:
  required: False
  type: list
  schema:
    type: dict
    schema: account
users:
  required: False
  type: list
  schema:
    type: dict
    schema: user
groups:
  required: False
  type: list
  schema:
    type: dict
    schema: group
delegations:
  required: False
  type: list
  schema:
    type: dict
    schema: delegation
local_users:
  required: False
  type: list
  schema:
    type: dict
    schema: local_user
custom_policies:
  required: False
  type: list
  schema:
    type: dict
    schema: custom_policy
"""


# Schema for validating the fully accumulate spec object.  This is where we
# ensure all required keys are present.  But we do not need to check sub
# schema, as that is done during spec_file validation.
#
SPEC_SCHEMA = """
master_account_id:
  required: True
  type: string
auth_account_id:
  required: True
  type: string
default_domain:
  required: True
  type: string
default_sc_policy:
  required: True
  type: string
default_ou:
  required: True
  type: string
default_path:
  required: True
  type: string
default_smtp_server:
  required: True
  type: string
org_admin_team:
  required: True
  type: string
organizational_units:
  required: True
  type: list
sc_policies:
  required: True
  type: list
teams:
  required: True
  type: list
accounts:
  required: True
  type: list
users:
  required: True
  type: list
groups:
  required: True
  type: list
delegations:
  required: True
  type: list
local_users:
  required: True
  type: list
custom_policies:
  required: True
  type: list
"""


ORGANIZATIONAL_UNIT_SCHEMA = """
Name:
  required: True
  type: string
Accounts:
  required: False
  nullable: True
  type: list
  schema:
    type: string
Child_OU:
  required: False
  nullable: True
  type: list
  schema:
    type: dict
    schema: organizational_unit
SC_Policies:
  required: False
  nullable: True
  type: list
  schema:
    type: string
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

SC_POLICY_SCHEMA = """
Name:
  required: True
  type: string
Description:
  required: False
  type: string
Effect:
  required: False
  type: string
  allowed:
  - Allow
  - Deny
Actions:
  required: False
  type: list
  schema:
    type: string
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

TEAM_SCHEMA = """
Name:
  required: True
  type: string
Description:
  required: True
  type: string
BusinessContacts:
  required: True
  type: list
  schema:
    type: string
TechnicalContacts:
  required: True
  type: list
  schema:
    type: string
"""

ACCOUNT_SCHEMA = """
Name:
  required: True
  type: string
Email:
  required: False
  type: string
Team:
  required: True
  type: string
Alias:
  required: False
  type: string
"""

USER_SCHEMA = """
Name:
  required: True
  type: string
Team:
  required: True
  type: string
Email:
  required: True
  type: string
Path:
  required: False
  type: string
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

GROUP_SCHEMA = """
Name:
  required: True
  type: string
Path:
  required: False
  type: string
Members:
  required: False
  nullable: True
  anyof:
  - type: string
    allowed:  
    - ALL
  - type: list
    schema:
      type: string
ExcludeMembers:
  required: False
  nullable: True
  type: list
  schema:
    type: string
Policies:
  required: False
  nullable: True
  type: list
  schema:
    type: string
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

LOCAL_USER_SCHEMA = """
Name:
  required: True
  type: string
Description:
  required: False
  type: string
Team:
  required: True
  type: string
Path:
  required: False
  type: string
#AuthMethod:
#  required: False
#  type: string
Account:
  required: True
  anyof:
  - type: string
    allowed:  
    - ALL
  - type: list
    schema:
      type: string
ExcludeAccounts:
  required: False
  type: list
  schema:
    type: string
#TrustedGroup:
#  required: False
#  type: string
Policies:
  required: False
  type: list
  schema:
    type: string
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

DELEGATION_SCHEMA = """
RoleName:
  required: True
  type: string
Description:
  required: False
  type: string
TrustingAccount:
  required: True
  anyof:
  - type: string
    allowed:  
    - ALL
  - type: list
    schema:
      type: string
ExcludeAccounts:
  required: False
  type: list
  schema:
    type: string
TrustedGroup:
  required: False
  type: string
TrustedAccount:
  required: False
  type: string
RequireMFA:
  required: False
  type: boolean
Policies:
  required: False
  type: list
  schema:
    type: string
Path:
  required: False
  type: string
Duration:
  required: False
  type: integer
  min: 3600
  max: 43200
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""

CUSTOM_POLICY_SCHEMA = """
PolicyName:
  required: True
  type: string
Description:
  required: False
  type: string
Statement:
  required: True
  anyof:
  - type: string
  - type: list
    schema:
      type: dict
Ensure:
  required: False
  type: string
  allowed:
  - present
  - absent
"""


def file_validator(log):
    schema_registry.add('organizational_unit', yaml.load(ORGANIZATIONAL_UNIT_SCHEMA))
    schema_registry.add('sc_policy', yaml.load(SC_POLICY_SCHEMA))
    schema_registry.add('team', yaml.load(TEAM_SCHEMA))
    schema_registry.add('account', yaml.load(ACCOUNT_SCHEMA))
    schema_registry.add('user', yaml.load(USER_SCHEMA))
    schema_registry.add('group', yaml.load(GROUP_SCHEMA))
    schema_registry.add('local_user', yaml.load(LOCAL_USER_SCHEMA))
    schema_registry.add('delegation', yaml.load(DELEGATION_SCHEMA))
    schema_registry.add('custom_policy', yaml.load(CUSTOM_POLICY_SCHEMA))
    log.debug("adding subschema to schema_registry: {}".format(
            schema_registry.all().keys()))
    vfile = Validator(yaml.load(SPEC_FILE_SCHEMA))
    log.debug("file_validator_schema: {}".format(vfile.schema))
    return vfile


def spec_validator(log):
    vspec = Validator(yaml.load(SPEC_SCHEMA))
    log.debug("spec_validator_schema: {}".format(vspec.schema))
    return vspec



import sys
import os
import yaml

import boto3
from botocore.exceptions import ClientError
from cerberus import Validator, schema_registry

from awsorgs.utils import *
from awsorgs.validator import file_validator, spec_validator

# Spec parser defaults
DEFAULT_CONFIG_FILE = '~/.awsorgs/config.yaml'
DEFAULT_SPEC_DIR = '~/.awsorgs/spec.d'



def scan_config_file(log, args):
    if args['--config']:
        config_file = args['--config']
    else:
        config_file = DEFAULT_CONFIG_FILE
    config_file = os.path.expanduser(config_file)
    if not os.path.isfile(config_file):
        log.error("config_file not found: {}".format(config_file))
        return None
    log.debug("loading config file: {}".format(config_file))
    with open(config_file) as f:
        try:
            config = yaml.load(f.read())
        except (yaml.scanner.ScannerError, UnicodeDecodeError):
            log.error("{} not a valid yaml file".format(config_file))
            return None
        except Exception as e:
            log.error("cant load config_file '{}': {}".format(config_file, e))
            return None
    log.debug("config: {}".format(config))
    return config


def get_master_account_id(log, args, config):
    """
    Determine the Org Master account id.  Try in order:
    cli option, config file, client.describe_organization()
    """
    if args['--master-account-id']:
        master_account_id = args['--master-account-id']
    else:
        master_account_id = config.get('master_account_id')
    if master_account_id:
        if not valid_account_id(log, master_account_id):
            log.critical("config option 'master_account_id' is not valid account Id")
            sys.exit(1)
    else:
        log.debug("'master_account_id' not set in config_file or as cli option")
        try:
            master_account_id = boto3.client('organizations'
                    ).describe_organization()['Organization']['MasterAccountId']
        except ClientError as e:
            log.critical("can not determine master_account_id: {}".format(e))
            sys.exit(1)
    log.debug("master_account_id: %s" % master_account_id)
    return master_account_id


def get_spec_dir(log, args, config):
    """
    Determine the spec directory.  Try in order:
    cli option, config file, DEFAULT_SPEC_DIR.
    """
    if '--spec-dir' in args and args['--spec-dir']:
        spec_dir = args['--spec-dir']
    elif config['spec_dir']:
        spec_dir = config['spec_dir']
    else:
        spec_dir = DEFAULT_SPEC_DIR
    spec_dir = os.path.expanduser(spec_dir)
    log.debug("spec_dir: %s" % spec_dir)
    return spec_dir


def load_config(log, args):
    """
    Assemble config options from various sources: cli options, config_file 
    params, defaults, etc., and merge them into 'args' dict.
    When we are done we should have found all of the following:

    master_account_id
    org_access_role
    spec_dir (except when handling reports)
    auth_account_id (except when called by awsorgs)
    """
    config = scan_config_file(log, args)
    args['--master-account-id'] = get_master_account_id(log, args, config)
    args['--spec-dir'] = get_spec_dir(log, args, config)
    if not args['--org-access-role']:
        args['--org-access-role'] =  config.get('org_access_role')
    if not args['--auth-account-id']:
        args['--auth-account-id'] =  config.get('auth_account_id')
    return args


def validate_spec_file(log, spec_file, validator, errors):
    with open(spec_file) as f:
        try:
            spec_from_file = yaml.load(f.read())
        except (yaml.scanner.ScannerError, UnicodeDecodeError):
            log.warn("{} not a valid yaml file. skipping".format(spec_file))
            return (None, errors)
        except Exception as e:
            log.error("cant load spec_file '{}': {}".format(spec_file, e))
            return (None, errors)
    if validator.validate(spec_from_file):
        return (spec_from_file, errors)
    else:
        log.error("schema validation failed for spec_file: {}".format(spec_file))
        log.debug("validator errors:\n{}".format(yamlfmt(validator.errors)))
        errors += 1
        return (None, errors)


def validate_spec(log, args):
    """
    Load all spec files in spec_dir and validate against spec schema
    """

    # validate spec_files
    spec_dir = args['--spec-dir']
    if not os.path.isdir(spec_dir):
        log.error("spec_dir not found or not a directory: {}".format(spec_dir))
        sys.exit(1)
    validator = file_validator(log)
    spec_object = {}
    errors = 0
    for dirpath, dirnames, filenames in os.walk(spec_dir, topdown = True):
        dirnames[:] = [d for d in dirnames if not d.startswith('.')]
        for f in filenames:
            log.debug("considering file {}".format(f))
            spec_from_file, errors = validate_spec_file(log,
                    os.path.join(dirpath, f), validator, errors)
            if spec_from_file:
                spec_object.update(spec_from_file)
    if errors:
        log.critical("schema validation failed for {} spec files. run in debug mode for details".format(errors))
        sys.exit(1)
    log.debug("spec_object:\n{}".format(yamlfmt(spec_object)))

    # validate aggregated spec_object
    validator = spec_validator(log)
    if validator.validate(spec_object):
        log.debug("spec_object validation succeeded")
        return spec_object
    else:
        log.critical("spec_object validation failed:\n{}".format(
                yamlfmt(validator.errors)))
        sys.exit(1)




#!/usr/bin/env python


"""Manage users, group, and roles for cross account authentication in an
AWS Organization.

Usage:
  awsauth (users|delegations|local-users|report) [--config FILE]
                                                 [--spec-dir PATH] 
                                                 [--master-account-id ID]
                                                 [--auth-account-id ID]
                                                 [--org-access-role ROLE]
                                                 [--disable-expired]
                                                 [--opt-ttl HOURS]
                                                 [--users --roles --credentials]
                                                 [--account NAME] [--full]
                                                 [--exec] [-q] [-d|-dd]
  awsauth (--help|--version)

Modes of operation:
  users         Provision users, groups and group membership.
  delegation    Provision policies and roles for cross account access.
  local-users   Provision local IAM users and policies in accounts.
  report        Display provisioned resources.

Options:
  -h, --help                Show this help message and exit.
  -V, --version             Display version info and exit.
  --config FILE             AWS Org config file in yaml format.
  --spec-dir PATH           Location of AWS Org specification file directory.
  --master-account-id ID    AWS account Id of the Org master account.    
  --auth-account-id ID      AWS account Id of the authentication account.
  --org-access-role ROLE    IAM role for traversing accounts in the Org.
  --exec                    Execute proposed changes to AWS accounts.
  -q, --quiet               Repress log output.
  -d, --debug               Increase log level to 'DEBUG'.
  -dd                       Include botocore and boto3 logs in log stream.

  users options:
  --disable-expired         Delete profile if one-time-password
                            exceeds --opt-ttl.
  --opt-ttl HOURS           One-time-password time to live in hours
                            [default: 24].
  report options:
  --users                   Print user and groups report.
  --roles                   Print roles and custom policies report.
  --credentials             Print IAM credentials report.
  --full                    Print full details in reports.
  --account NAME            Just report for a single named account.

"""

import os
import sys
import yaml
import json

import boto3
from botocore.exceptions import ClientError
from docopt import docopt

import awsorgs
from awsorgs.utils import *
from awsorgs.spec import *
from awsorgs.loginprofile import *
from awsorgs.reports import *


def expire_users(log, args, deployed, auth_spec, credentials):
    """
    Delete login profile for any users whose one-time-password has expired
    """
    for name in [u['UserName'] for u in deployed['users']]:
        user = validate_user(name, credentials)
        if user:
            login_profile = validate_login_profile(user)
            if login_profile and onetime_passwd_expired(log, user, login_profile,
                    int(args['--opt-ttl'])):
                log.info('deleting login profile for user %s' % user.name)
                if args['--exec']:
                    login_profile.delete()


def delete_user(user):
    """
    Strip user attributes and delete user

    :param: user
    :type:  boto3 iam User resource object
    """
    try:
        user.load()
    except user.meta.client.exceptions.NoSuchEntityException:
        return
    for x in user.access_keys.all():
        x.delete()
    for x in user.attached_policies.all():
        x.detach_user(UserName=user.name)
    for x in user.groups.all():
        x.remove_user(UserName=user.name)
    for x in user.mfa_devices.all():
        x.disassociate()
    for x in user.policies.all():
        x.delete()
    for x in user.signing_certificates.all():
        x.delete()
    profile = user.LoginProfile()
    try:
        profile.load()
        profile.delete()
    except profile.meta.client.exceptions.NoSuchEntityException:
        pass
    user.delete()


def delete_policy(policy):
    """
    Delete IAM policy.
    
    Args:
        policy (obj): boto3 IAM resource object
    """
    if policy.attachment_count > 0:
        for group in policy.attached_groups.all():
            policy.detach_group(GroupName=group.name)
        for user in policy.attached_users.all():
            policy.detach_user(UserName=user.name)
        for role in policy.attached_roles.all():
            policy.detach_role(RoleName=role.name)
    for version in policy.versions.all():
        if not version.is_default_version:
            version.delete()
    policy.delete()


def create_users(credentials, args, log, deployed, auth_spec):
    """
    Manage IAM users based on user specification
    """
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)
    for u_spec in auth_spec['users']:
        path = munge_path(auth_spec['default_path'], u_spec)
        deployed_user = lookup(deployed['users'], 'UserName', u_spec['Name'])
        if deployed_user:
            user = iam_resource.User(u_spec['Name'])
            # delete user
            if ensure_absent(u_spec):
                log.info("Deleting user '%s'" % user.name)
                if args['--exec']:
                    delete_user(user)
            # update user
            elif user.path != path:
                log.info("Updating path on user '%s'" % u_spec['Name'])
                if args['--exec']:
                    user.update(NewPath=path)
        # create new user
        elif not ensure_absent(u_spec):
            log.info("Creating user '%s'" % u_spec['Name'])
            if args['--exec']:
                response = iam_client.create_user(UserName=u_spec['Name'], Path=path)
                log.info(response['User']['Arn'])
                deployed['users'].append(response['User'])


def create_groups(credentials, args, log, deployed, auth_spec):
    """
    Manage IAM groups based on group specification
    """
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)
    for g_spec in auth_spec['groups']:
        path = munge_path(auth_spec['default_path'], g_spec)
        deployed_group = lookup(deployed['groups'], 'GroupName', g_spec['Name'])
        if deployed_group:
            group = iam_resource.Group(g_spec['Name'])
            # delete group?
            if ensure_absent(g_spec):
                # check if group has users
                if list(group.users.all()):
                    log.error("Can not delete group '%s'. Still contains users"
                             % g_spec['Name'])
                else:
                    log.info("Deleting group '%s'" % g_spec['Name'])
                    if args['--exec']:
                        for policy in group.policies.all():
                            policy.delete()
                        for policy in group.attached_policies.all():
                            policy.detach_group(GroupName=g_spec['Name'])
                        group.delete()
                        deployed['groups'].remove(deployed_group)
            # update group?
            elif group.path != path:
                log.info("Updating path on group '%s'" % g_spec['Name'])
                if args['--exec']:
                    group.update(NewPath=path)
        # create group
        elif not ensure_absent(g_spec):
            log.info("Creating group '%s'" % g_spec['Name'])
            if args['--exec']:
                response = iam_client.create_group(
                        GroupName=g_spec['Name'], Path=path)
                log.info(response['Group']['Arn'])
                deployed['groups'].append(response['Group'])


def manage_group_members(credentials, args, log, deployed, auth_spec):
    """
    Populate users into groups based on group specification.
    """
    iam_resource = boto3.resource('iam', **credentials)
    for g_spec in auth_spec['groups']:
        if lookup(deployed['groups'], 'GroupName', g_spec['Name']):
            group = iam_resource.Group(g_spec['Name'])
            current_members = [user.name for user in group.users.all()] 
            # build list of specified group members
            spec_members = []
            if 'Members' in g_spec and g_spec['Members']:
                if g_spec['Members'] == 'ALL':
                    # all managed users except when user ensure: absent
                    spec_members = [user['Name'] for user in auth_spec['users']
                            if not ensure_absent(user)]
                    if 'ExcludeMembers' in g_spec and g_spec['ExcludeMembers']:
                        spec_members = [user for user in spec_members
                                if user not in g_spec['ExcludeMembers']]
                else:
                    # just specified members
                    for username in g_spec['Members']:
                        u_spec = lookup(auth_spec['users'], 'Name', username)
                        # not a managed user?
                        if not u_spec:
                            log.error("User '%s' not in auth_spec['users']. "
                                    "Can not add user to group '%s'" %
                                    (username, g_spec['Name']))
                        # managed but absent?
                        elif ensure_absent(u_spec):
                            log.error("User '%s' is specified 'absent' in "
                                    "auth_spec['users']. Can not add user "
                                    "to group '%s'" % 
                                    (username, g_spec['Name']))
                        else:
                            spec_members.append(username)
            # ensure all specified members are in group
            if not ensure_absent(g_spec):
                for username in spec_members:
                    if username not in current_members:
                        log.info("Adding user '%s' to group '%s'" %
                                (username, g_spec['Name']))
                        if args['--exec']:
                            group.add_user(UserName=username)
            # ensure no unspecified members are in group
            for username in current_members:
                if username not in spec_members:
                    log.info("Removing user '%s' from group '%s'" %
                            (username, g_spec['Name']))
                    if args['--exec']:
                        group.remove_user(UserName=username)


def manage_group_policies(credentials, args, log, deployed, auth_spec):
    """
    Attach managed policies to groups based on group specification
    """
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)
    auth_account = lookup(deployed['accounts'], 'Id',
            auth_spec['auth_account_id'], 'Name')
    log.debug("auth account: '%s'" % auth_account)
    for g_spec in auth_spec['groups']:
        if (lookup(deployed['groups'], 'GroupName', g_spec['Name'])
                and not ensure_absent(g_spec)):
            log.debug("processing group spec for '%s':\n%s" % (g_spec['Name'], g_spec))
            group = iam_resource.Group(g_spec['Name'])
            attached_policies = [p.policy_name for p in list(group.attached_policies.all())]
            log.debug("attached policies: '%s'" % attached_policies)
            if not 'Policies' in g_spec or g_spec['Policies'] is None:
                g_spec['Policies'] = []
            if g_spec['Policies']:
                log.debug("specified policies: '%s'" % g_spec['Policies'])
                # attach missing policies
                for policy_name in g_spec['Policies']:
                    if not policy_name in attached_policies:
                        policy_arn = get_policy_arn(iam_client, policy_name)
                        if policy_arn is None:
                            policy_arn = manage_custom_policy(iam_client, auth_account,
                                    policy_name, args, log, auth_spec)
                        log.debug("policy Arn for '%s': %s" % (policy_name, policy_arn))
                        log.info("Attaching policy '%s' to group '%s' in "
                                "account '%s'" % (policy_name, g_spec['Name'],
                                auth_account))
                        if args['--exec']:
                            group.attach_policy(PolicyArn=policy_arn)
                    # update custom policy
                    elif lookup(auth_spec['custom_policies'], 'PolicyName', policy_name):
                        manage_custom_policy(iam_client, auth_account, policy_name,
                                args, log, auth_spec)
            # datach obsolete policies
            for policy_name in attached_policies:
                if not policy_name in g_spec['Policies']:
                    policy_arn = get_policy_arn(iam_client, policy_name)
                    log.info("Detaching policy '%s' from group '%s' in "
                            "account '%s'" % (policy_name, g_spec['Name'],
                            auth_account))
                    if args['--exec']:
                        group.detach_policy(PolicyArn=policy_arn)


def get_policy_arn(iam_client, policy_name):
    """
    Return the policy arn of the named IAM policy in an account.
    """
    aws_policies = get_iam_objects(iam_client.list_policies, 'Policies')
    return lookup(aws_policies, 'PolicyName', policy_name, 'Arn')


def manage_custom_policy(iam_client, account_name, policy_name, args, log, auth_spec):
    """
    Create or update a custom IAM policy in an account based on a 
    policy specification.  Returns the policy arn.
    """
    log.debug("account: '{}', policy_name: '{}'".format(account_name, policy_name))
    p_spec = lookup(auth_spec['custom_policies'], 'PolicyName', policy_name)
    if not p_spec:
        log.error("Custom Policy spec for '%s' not found in auth-spec." % policy_name)
        log.error("Policy creation failed.")
        return None
    policy_doc = dict(Version='2012-10-17', Statement=p_spec['Statement'])

    # check if custom policy exists
    custom_policies = get_iam_objects(
            iam_client.list_policies, 'Policies', dict(Scope='Local'))
    log.debug("account: '%s', custom policies: '%s'" % (
            account_name,
            [p['Arn'] for p in custom_policies]))
    policy = lookup(custom_policies, 'PolicyName', policy_name)
    if not policy:
        log.info("Creating custom policy '%s' in account '%s':\n%s" %
                (policy_name, account_name, yamlfmt(policy_doc)))
        if args['--exec']:
            return iam_client.create_policy(
                PolicyName=policy_name,
                Path=munge_path(auth_spec['default_path'], p_spec),
                Description=p_spec['Description'],
                PolicyDocument=json.dumps(policy_doc),
            )['Policy']['Arn']
        return None

    # check if custom policy needs updating
    else:
        current_doc = iam_client.get_policy_version(
                PolicyArn=policy['Arn'],
                VersionId=policy['DefaultVersionId']
                )['PolicyVersion']['Document']
        log.debug("account: '%s', policy_doc: %s" % (account_name, policy_doc))
        log.debug("account: '%s', current_doc: %s" % (account_name, current_doc))

        # compare each statement as dict
        update_required = False
        if current_doc['Statement'] != policy_doc['Statement']:
            update_required = True
            log.debug('account: %s, update_required: %s' %
                    (account_name, update_required))

        # update policy and set as default version
        if update_required:
            log.info("Updating custom policy '%s' in account '%s':\n%s" % (
                    policy_name,
                    account_name, 
                    string_differ(yamlfmt(current_doc), yamlfmt(policy_doc))))
            if args['--exec']:
                log.debug("check for non-default policy versions for '%s'" % policy_name)
                for v in iam_client.list_policy_versions(
                        PolicyArn=policy['Arn'])['Versions']:
                    if not v['IsDefaultVersion']:
                        log.info("Deleting non-default policy version '%s' for "
                                "policy '%s' in account '%s'" %
                                (v['VersionId'], policy_name, account_name))
                        iam_client.delete_policy_version(
                                PolicyArn=policy['Arn'],
                                VersionId=v['VersionId'])
                iam_client.create_policy_version(
                        PolicyArn=policy['Arn'],
                        PolicyDocument=json.dumps(policy_doc),
                        SetAsDefault=True)
        return policy['Arn']


def build_role_arn(account_id, d_spec, auth_spec):
    return 'arn:aws:iam::{}:role{}{}'.format(
        account_id,
        munge_path(auth_spec['default_path'], d_spec),
        d_spec['RoleName']
    )


def build_resource_list(log, deployed_accounts, d_spec, auth_spec, account_list):
    resource = []
    for account in account_list:
        account_id = lookup(deployed_accounts, 'Name', account, 'Id')
        if account_id is not None:
            resource.append(build_role_arn(account_id, d_spec, auth_spec))
        else:
            log.warn('Account {} not found in deployed accounts'.format(account))
    return resource


def assemble_assume_role_policy_document(resource, effect):
    statement = dict(
        Effect=effect,
        Action='sts:AssumeRole',
        Resource=resource,
    )
    return dict(Version='2012-10-17', Statement=[statement])


def create_group_policy(args, log, group, account, policy_name, policy_doc):
    log.info("Creating assume role policy '{}' for group '{}' in account '{}':\n{}".format(
        policy_name, 
        group.name,
        account, 
        yamlfmt(policy_doc),
    ))
    if args['--exec']:
        group.create_policy(
            PolicyName=policy_name,
            PolicyDocument=json.dumps(policy_doc),
        )


def update_group_policy(args, log, group, account, policy_name, policy_doc):
    log.info("Updating policy '{}' for group '{}' in account '{}':\n{}".format(
        policy_name, 
        group.name,
        account,
        string_differ(
            yamlfmt(group.Policy(policy_name).policy_document), 
            yamlfmt(policy_doc),
        ),
    ))
    if args['--exec']:
        group.Policy(policy_name).put(PolicyDocument=json.dumps(policy_doc))


def manage_group_policy(args, log, group, account, policy_name, policy_doc, group_policies):
    if not policy_name in group_policies:
        create_group_policy(args, log, group, account, policy_name, policy_doc)
    elif group.Policy(policy_name).policy_document != policy_doc:
        update_group_policy(args, log, group, account, policy_name, policy_doc)


def delete_group_policy(args, log, group, account, policy_name):
    log.info("Deleting assume role group policy '{}' from group '{}' in account '{}'".format(
        policy_name,
        group.name,
        account,
    ))
    if args['--exec']:
        group.Policy(policy_name).delete()


def delete_obsolete_group_policy(args, log, group, account, policy_name, managed_policies):
    if policy_name not in managed_policies:
        log.info("Deleting obsolete policy '{}' from group '{}' in account '{}'".format(
            policy_name,
            group.name,
            account,
        ))
        if args['--exec']:
            group.Policy(policy_name).delete()


def set_group_assume_role_policies(args, log, deployed, auth_spec, d_spec):
    """
    Assign and manage assume role trust policies on IAM groups in
    Auth account.
    """
    log.debug('role: %s' % d_spec['RoleName'])
    credentials = get_assume_role_credentials(
        args['--auth-account-id'],
        args['--org-access-role'],
    )
    iam_resource = boto3.resource('iam', **credentials)
    auth_account = lookup(deployed['accounts'], 'Id', auth_spec['auth_account_id'], 'Name')
    managed_policies = []
    if lookup(deployed['groups'], 'GroupName', d_spec['TrustedGroup']):
        group = iam_resource.Group(d_spec['TrustedGroup'])
        group.load()
    else:
        log.error(
            "Can not manage assume role policy for delegation role '{}' in group '{}'. "
            "Group not found in auth account '{}'".format(
                d_spec['RoleName'],
                d_spec['TrustedGroup'],
                auth_account,
            )
        )
        return

    # make list of existing group policies which match this role name
    group_policies = [
        p.policy_name for p in list(group.policies.all())
        if d_spec['RoleName'] in p.policy_name.split('-')
    ]

    # test if delegation should be deleted
    if ensure_absent(d_spec): 
        for policy_name in group_policies:
            delete_group_policy(args, log, group, auth_account, policy_name)
        return

    # handle trusting accounts
    if d_spec['TrustingAccount'] == 'ALL':
        resource = build_role_arn('*', d_spec, auth_spec)
    else:
        resource = build_resource_list(
            log, deployed['accounts'], d_spec, auth_spec, d_spec['TrustingAccount']
        )
    policy_doc = assemble_assume_role_policy_document(resource, 'Allow')
    policy_name = "AllowAssumeRole-{}".format(d_spec['RoleName'])
    manage_group_policy(
        args, log, group, auth_account, policy_name, policy_doc, group_policies
    )
    managed_policies.append(policy_name)

    # handle excluded accounts
    if 'ExcludeAccounts' in d_spec and d_spec['ExcludeAccounts'] is not None:
        resource = build_resource_list(
            log, deployed['accounts'], d_spec, auth_spec, d_spec['ExcludeAccounts']
        )
        policy_doc = assemble_assume_role_policy_document(resource, 'Deny')
        policy_name = "DenyAssumeRole-{}".format(d_spec['RoleName'])
        manage_group_policy(
            args, log, group, auth_account, policy_name, policy_doc, group_policies
        )
        managed_policies.append(policy_name)

    # purge any policies for this role that are no longer being managed
    for policy_name in group_policies:
        delete_obsolete_group_policy(args, log, group, auth_account, policy_name, managed_policies)


def manage_local_user_in_accounts(
            account, args, log, auth_spec, deployed, accounts, lu_spec):
    """
    Create and manage a local user in an account per user specification.
    """

    account_name = account['Name']
    log.debug('account: %s, local user: %s' % (account_name, lu_spec['Name']))
    path_spec = munge_path(auth_spec['default_path'], lu_spec)
    credentials = get_assume_role_credentials(account['Id'], args['--org-access-role'])
    if isinstance(credentials, RuntimeError):
        log.error(credentials)
        return
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)

    # get iam user object.
    user = iam_resource.User(lu_spec['Name'])
    try:
        user.load()
    except user.meta.client.exceptions.NoSuchEntityException:
        user_exists = False
    else:
        user_exists = True
        log.debug('account: %s, local user exists: %s' % (account_name, user.arn))

    # check for unmanaged user in account
    if user_exists:
        if not user.path.startswith('/' + auth_spec['default_path']):
            log.error(
                    "Can not manage local user '%s' in account '%s'. "
                    " Unmanaged user with the same name already exists: %s" % 
                    (user.name, account_name, user.arn))
            return

    # check if local user should not exist
    if account_name not in accounts or ensure_absent(lu_spec):
        if user_exists:
            log.info("Deleting local user '%s' from account '%s'" %
                    (user.name, account_name))
            if args['--exec']:
                delete_user(user)
        return

    # create local user and attach policies
    if not user_exists:
        log.info("Creating local user '%s' in account '%s'" %
                (lu_spec['Name'], account_name))
        if args['--exec']:
            user.create(Path=path_spec)
            if 'Policies' in lu_spec and lu_spec['Policies']:
                user.load()
                for policy_name in lu_spec['Policies']:
                    policy_arn = get_policy_arn(iam_client, policy_name)
                    if policy_arn is None:
                        policy_arn = manage_custom_policy(iam_client, account_name,
                                policy_name, args, log, auth_spec)
                    log.info("Attaching policy '%s' to local user '%s' "
                            "in account '%s'" %
                            (policy_name, user.name, account_name))
                    if args['--exec'] and policy_arn:
                        user.attach_policy(PolicyArn=policy_arn)
    else:
        # validate path
        if user.path != path_spec:
            log.info("Updating path for local user '%s'" % user.arn)
            if args['--exec']:
                # hack around bug in boto3
                try:
                    user.update(NewPath=path_spec)
                except AttributeError as e:
                    log.debug('boto3 error when calling user.update(): %s' % e)

        # manage policy attachments
        attached_policies = [p.policy_name for p in list(user.attached_policies.all())]
        for policy_name in lu_spec['Policies']:
            if not policy_name in attached_policies:
                policy_arn = get_policy_arn(iam_client, policy_name)
                if policy_arn is None:
                    policy_arn = manage_custom_policy(iam_client, account_name,
                            policy_name, args, log, auth_spec)
                log.info("Attaching policy '%s' to local user '%s' in account '%s'" %
                        (policy_name, user.name, account_name))
                if args['--exec'] and policy_arn:
                    user.attach_policy(PolicyArn=policy_arn)
            elif lookup(auth_spec['custom_policies'], 'PolicyName',policy_name):
                manage_custom_policy(iam_client, account_name, policy_name, args,
                        log, auth_spec)
        # datach obsolete policies
        for policy_name in attached_policies:
            if not policy_name in lu_spec['Policies']:
                policy_arn = get_policy_arn(iam_client, policy_name)
                log.info("Detaching policy '%s' from local user '%s' in account '%s'" %
                        (policy_name, user.name, account_name))
                if args['--exec'] and policy_arn:
                    user.detach_policy(PolicyArn=policy_arn)


def manage_local_users(lu_spec, args, log, deployed, auth_spec):
    """
    Create and manage local IAM users in specified accounts and 
    attach policies to users based on local_user specifications.
    """
    log.debug('considering %s' % lu_spec['Name'])
    # munge accounts list
    if lu_spec['Account'] == 'ALL':
        accounts = [a['Name'] for a in deployed['accounts']]
        if 'ExcludeAccounts' in lu_spec and lu_spec['ExcludeAccounts']:
            accounts = [a for a in accounts
                    if a not in lu_spec['ExcludeAccounts']]
    else:
        accounts = lu_spec['Account']
    for account_name in accounts:
        if not lookup(deployed['accounts'], 'Name', account_name):
            log.error("Can not manage local user '%s' in account "
                    "'%s'.  Account '%s' not found in Organization" %
                    (lu_spec['Name'], account_name, account_name))
            accounts.remove(account_name)
    # run manage_local_user_in_accounts() task in thread pool
    queue_threads(log, deployed['accounts'], manage_local_user_in_accounts,
            f_args=(args, log, auth_spec, deployed, accounts, lu_spec))


def manage_delegation_role(account, args, log, auth_spec, deployed,
            trusting_accounts, d_spec):
    """
    Create and manage a cross account access delegetion role in an
    account based on delegetion specification.
    """
    account_name = account['Name']
    log.debug('account: %s, role: %s' % (account_name, d_spec['RoleName']))
    credentials = get_assume_role_credentials(
            account['Id'],
            args['--org-access-role'])
    if isinstance(credentials, RuntimeError):
        log.error(credentials)
        return
    iam_client = boto3.client('iam', **credentials)
    iam_resource = boto3.resource('iam', **credentials)
    role = iam_resource.Role(d_spec['RoleName'])

    # check if role should not exist
    if account_name not in trusting_accounts or ensure_absent(d_spec):
        try:
            role.load()
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntity':
                return
            else:
                raise e
        except:
            raise
        # delete delegation role
        log.info("Deleting role '%s' from account '%s'" %
                (d_spec['RoleName'], account_name))
        if args['--exec']:
            for p in list(role.attached_policies.all()):
                role.detach_policy(PolicyArn=p.arn)
            role.delete()
        return

    # else: assemble assume role policy document for delegation role
    if 'TrustedAccount' in d_spec and d_spec['TrustedAccount']:
        trusted_account = lookup(deployed['accounts'], 'Name',
                d_spec['TrustedAccount'], 'Id')
    else:
        trusted_account = auth_spec['auth_account_id']
    principal = "arn:aws:iam::%s:root" % trusted_account
    statement = dict(
            Effect='Allow',
            Principal=dict(AWS=principal),
            Action='sts:AssumeRole')
    mfa = True
    if 'RequireMFA' in d_spec and d_spec['RequireMFA'] == False:
        mfa = False
    if mfa:
        statement['Condition'] = {
                'Bool':{'aws:MultiFactorAuthPresent':'true'}}
    policy_doc = dict(Version='2012-10-17', Statement=[statement])

    # munge session duration
    if not 'Duration' in d_spec:
        d_spec['Duration'] = 3600

    # get iam role object.  create role if it does not exist (i.e. won't load)
    try:
        role.load()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            log.info("Creating role '%s' in account '%s'" %
                    (d_spec['RoleName'], account_name))
            if args['--exec']:
                iam_client.create_role(
                        Description=d_spec['Description'],
                        Path=munge_path(auth_spec['default_path'], d_spec),
                        RoleName=d_spec['RoleName'],
                        MaxSessionDuration=d_spec['Duration'],
                        AssumeRolePolicyDocument=json.dumps(policy_doc))
                if 'Policies' in d_spec and d_spec['Policies']:
                    role.load()
                    for policy_name in d_spec['Policies']:
                        policy_arn = get_policy_arn(iam_client, policy_name)
                        if policy_arn is None:
                            policy_arn = manage_custom_policy(iam_client, account_name,
                                    policy_name, args, log, auth_spec)
                        log.info("Attaching policy '%s' to role '%s' "
                                "in account '%s':\n%s" % (
                                        policy_name, 
                                        d_spec['RoleName'], 
                                        account_name,
                                        yamlfmt(policy_doc)))
                        if args['--exec'] and policy_arn:
                            role.attach_policy(PolicyArn=policy_arn)
            return

    # update delegation role if needed
    if role.assume_role_policy_document != policy_doc:
        log.info("Updating policy document in role '%s' in account '%s':\n%s" % (
                d_spec['RoleName'], 
                account_name,
                string_differ(
                        yamlfmt(role.assume_role_policy_document),
                        yamlfmt(policy_doc))))
        if args['--exec']:
            iam_client.update_assume_role_policy(
                RoleName=role.role_name,
                PolicyDocument=json.dumps(policy_doc))
    if role.description != d_spec['Description']:
        log.info("Updating description in role '%s' in account '%s'" %
                (d_spec['RoleName'], account_name))
        if args['--exec']:
            iam_client.update_role_description(
                RoleName=role.role_name,
                Description=d_spec['Description'])
    if role.max_session_duration != d_spec['Duration']:
        log.info("Updating max session duration in role '%s' in account '%s'" %
                (d_spec['RoleName'], account_name))
        if args['--exec']:
            iam_client.update_role(
                RoleName=role.role_name,
                MaxSessionDuration=d_spec['Duration'])

    # manage policy attachments
    attached_policies = [p.policy_name for p in list(role.attached_policies.all())]
    for policy_name in d_spec['Policies']:
        # attach missing policies
        if not policy_name in attached_policies:
            policy_arn = get_policy_arn(iam_client, policy_name)
            if policy_arn is None:
                policy_arn = manage_custom_policy(iam_client, account_name, policy_name,
                        args, log, auth_spec)
            log.info("Attaching policy '%s' to role '%s' in account '%s'" %
                    (policy_name, d_spec['RoleName'], account_name))
            if args['--exec'] and policy_arn:
                role.attach_policy(PolicyArn=policy_arn)
        elif lookup(auth_spec['custom_policies'], 'PolicyName',policy_name):
            manage_custom_policy(iam_client, account_name, policy_name,
                    args, log, auth_spec)
    for policy_name in attached_policies:
        # datach obsolete policies
        if not policy_name in d_spec['Policies']:
            policy_arn = get_policy_arn(iam_client, policy_name)
            log.info("Detaching policy '%s' from role '%s' in account '%s'" %
                    (policy_name, d_spec['RoleName'], account_name))
            if args['--exec'] and policy_arn:
                role.detach_policy(PolicyArn=policy_arn)


def manage_delegations(d_spec, args, log, deployed, auth_spec):
    """
    Create and manage cross account access delegations based on 
    delegation specifications.  Manages delegation roles in 
    trusting accounts and group policies in Auth (trusted) account.
    """
    log.debug('considering %s' % d_spec['RoleName'])
    if d_spec['RoleName'] == args['--org-access-role']:
        log.error("Refusing to manage delegation '%s'" % d_spec['RoleName'])
        return

    # munge trusting_accounts list
    if d_spec['TrustingAccount'] == 'ALL':
        trusting_accounts = [a['Name'] for a in deployed['accounts']]
        if 'ExcludeAccounts' in d_spec and d_spec['ExcludeAccounts']:
            trusting_accounts = [a for a in trusting_accounts
                    if a not in d_spec['ExcludeAccounts']]
    else:
        trusting_accounts = d_spec['TrustingAccount']
    for account_name in trusting_accounts:
        if not lookup(deployed['accounts'], 'Name', account_name):
            log.error("Can not manage delegation role '%s' in account "
                    "'%s'.  Account '%s' not found in Organization" %
                    (d_spec['RoleName'], account_name, account_name))
            trusting_accounts.remove(account_name)

    # is this a service role or a user role?
    if 'TrustedGroup' in d_spec and 'TrustedAccount' in d_spec:
        log.error("can not declare both 'TrustedGroup' or 'TrustedAccount' in "
                "delegation spec for role '%s'" % d_spec['RoleName'])
        return
    elif 'TrustedGroup' not in d_spec and 'TrustedAccount' not in d_spec:
        log.error("neither 'TrustedGroup' or 'TrustedAccount' declared in "
                "delegation spec for role '%s'" % d_spec['RoleName'])
        return
    elif 'TrustedAccount' in d_spec and d_spec['TrustedAccount']:
        # this is a service role. skip setting group policy
        pass
    else:
        # this is a user role. set group policies in Auth account
        set_group_assume_role_policies(args, log, deployed, auth_spec, d_spec)

    # run manage_delegation_role() task in thread pool
    queue_threads(log, deployed['accounts'], manage_delegation_role,
            f_args=(args, log, auth_spec, deployed, trusting_accounts, d_spec))


def main():
    args = docopt(__doc__, version=awsorgs.__version__)
    log = get_logger(args)
    log.debug("%s: args:\n%s" % (__name__, args))
    args = load_config(log, args)
    auth_spec = validate_spec(log, args)

    org_credentials = get_assume_role_credentials(
            args['--master-account-id'],
            args['--org-access-role'])
    if isinstance(org_credentials, RuntimeError):
        log.critical(org_credentials)
        sys.exit(1)
    org_client = boto3.client('organizations', **org_credentials)
    validate_master_id(org_client, auth_spec)

    auth_credentials = get_assume_role_credentials(
            args['--auth-account-id'],
            args['--org-access-role'])
    if isinstance(auth_credentials, RuntimeError):
        log.critical(auth_credentials)
        sys.exit(1)
    iam_client = boto3.client('iam', **auth_credentials)
    deployed = dict(
            users = get_iam_objects(iam_client.list_users, 'Users'),
            groups = get_iam_objects(iam_client.list_groups, 'Groups'),
            accounts = [a for a in scan_deployed_accounts(log, org_client)
                    if a['Status'] == 'ACTIVE'])

    if args['report']:
        if args['--account']:
            deployed['accounts'] = [lookup(
                deployed['accounts'], 'Name', args['--account']
            )]
        if args['--users']:
            report_maker(log, deployed['accounts'], args['--org-access-role'], 
                user_group_report, "IAM Users and Groups in all Org Accounts:",
                verbose=args['--full'],
            )
        if args['--roles']:
            report_maker(log, deployed['accounts'], args['--org-access-role'], 
                role_report, "IAM Roles and Custom Policies in all Org Accounts:",
                verbose=args['--full'],
            )
        if args['--credentials']:
            report_maker(log, deployed['accounts'], args['--org-access-role'], 
                credentials_report, "IAM Credentials Report in all Org Accounts:"
            )
        if not (args['--users'] or args['--credentials'] or args['--roles']):
            report_maker(log, deployed['accounts'], args['--org-access-role'], 
                account_authorization_report, "IAM Account Authorization:",
                verbose=args['--full'],
            )

    if args['users']:
        if args['--disable-expired']:
            expire_users(log, args, deployed, auth_spec, credentials)
        else:
            create_users(auth_credentials, args, log, deployed, auth_spec)
            create_groups(auth_credentials, args, log, deployed, auth_spec)
            manage_group_members(auth_credentials, args, log, deployed, auth_spec)
            manage_group_policies(auth_credentials, args, log, deployed, auth_spec)

    if args['delegations']:
        queue_threads(log, auth_spec['delegations'], manage_delegations,
            f_args=(args, log, deployed, auth_spec))

    if args['local-users']:
        queue_threads(log, auth_spec['local_users'], manage_local_users,
            f_args=(args, log, deployed, auth_spec))

if __name__ == "__main__":
    main()



#!/usr/bin/env python


"""Manage accounts in an AWS Organization.

Usage:
  awsaccounts (report|create|alias|invite) [--config FILE]
                                           [--spec-dir PATH] 
                                           [--master-account-id ID]
                                           [--auth-account-id ID]
                                           [--org-access-role ROLE]
                                           [--invited-account-id ID]
                                           [--exec] [-q] [-d|-dd]
  awsaccounts (--help|--version)

Modes of operation:
  report         Display organization status report.
  create         Create new accounts in AWS Org per specifation.
  alias          Set account alias for each account in Org per specifation.
  invite         Invite another account to join Org as a member account. 

Options:
  -h, --help                Show this help message and exit.
  -V, --version             Display version info and exit.
  -f, --config FILE         AWS Org config file in yaml format.
  --spec-dir PATH           Location of AWS Org specification file directory.
  --master-account-id ID    AWS account Id of the Org master account.    
  --auth-account-id ID      AWS account Id of the authentication account.
  --org-access-role ROLE    IAM role for traversing accounts in the Org.
  --invited-account-id ID   Id of account being invited to join Org.
                            Required when running in 'invite' mode.
  --exec                    Execute proposed changes to AWS accounts.
  --role ROLENAME           IAM role to use to access accounts.
  -q, --quiet               Repress log output.
  -d, --debug               Increase log level to 'DEBUG'.
  -dd                       Include botocore and boto3 logs in log stream.

"""


import yaml
import time

import boto3
import botocore
from botocore.exceptions import ClientError
from docopt import docopt

import awsorgs
from awsorgs.utils import *
from awsorgs.spec import *

S3_ACCOUNT_BUCKET = 'jjhsu-awsorgs-bucket'

def create_accounts(org_client, args, log, deployed_accounts, account_spec):
    """
    Compare deployed_accounts to list of accounts in the accounts spec.
    Create accounts not found in deployed_accounts.
    """
    for a_spec in account_spec['accounts']:
        if not lookup(deployed_accounts, 'Name', a_spec['Name']):
            # check if it is still being provisioned
            created_accounts = scan_created_accounts(log, org_client)
            if lookup(created_accounts, 'AccountName', a_spec['Name']):
                log.warn("New account '%s' is not yet available" % a_spec['Name'])
                break
            # create a new account
            if 'Email' in a_spec and a_spec['Email']:
                email_addr = a_spec['Email']
            else:
                email_addr = '%s@%s' % (a_spec['Name'], account_spec['default_domain'])
            log.info("Creating account '%s'" % (a_spec['Name']))
            log.debug('account email: %s' % email_addr)
            if args['--exec']:
                new_account = org_client.create_account(
                        AccountName=a_spec['Name'],
                        Email=email_addr)
                create_id = new_account['CreateAccountStatus']['Id']
                log.info("CreateAccountStatus Id: %s" % (create_id))
                # validate creation status
                counter = 0
                maxtries = 5
                while counter < maxtries:
                    creation = org_client.describe_create_account_status(
                            CreateAccountRequestId=create_id
                            )['CreateAccountStatus']
                    if creation['State'] == 'IN_PROGRESS':
                        time.sleep(5)
                        log.info("Account creation in progress for '%s'" %
                                a_spec['Name'])
                    elif creation['State'] == 'SUCCEEDED':
                        log.info("Account creation succeeded")
                        break
                    elif creation['State'] == 'FAILED':
                        log.error("Account creation failed: %s" %
                                creation['FailureReason'])
                        break
                    counter += 1
                if counter == maxtries and creation['State'] == 'IN_PROGRESS':
                     log.warn("Account creation still pending. Moving on!")


def set_account_alias(account, log, args, account_spec, role):
    """
    Set an alias on an account.  Use 'Alias' attribute from account spec
    if provided.  Otherwise set the alias to the account name.
    """
    if account['Status'] == 'ACTIVE':
        a_spec = lookup(account_spec['accounts'], 'Name', account['Name'])
        if a_spec and 'Alias' in a_spec:
            proposed_alias = a_spec['Alias']
        else:
            proposed_alias = account['Name'].lower()
        credentials = get_assume_role_credentials(
                account['Id'], args['--org-access-role'])
        if isinstance(credentials, RuntimeError):
            log.error(credentials)
            return
        else:
            iam_client = boto3.client('iam', **credentials)
        aliases = iam_client.list_account_aliases()['AccountAliases']
        log.debug('account_name: %s; aliases: %s' % (account['Name'], aliases))
        if not aliases:
            log.info("setting account alias to '%s' for account '%s'" %
                    (proposed_alias, account['Name']))
            if args['--exec']:
                try:
                    iam_client.create_account_alias(AccountAlias=proposed_alias)
                except Exception as e:
                    log.error(e)
        elif aliases[0] != proposed_alias:
            log.info("resetting account alias for account '%s' to '%s'; "
                    "previous alias was '%s'" %
                    (account['Name'], proposed_alias, aliases[0]))
            if args['--exec']:
                iam_client.delete_account_alias(AccountAlias=aliases[0])
                try:
                    iam_client.create_account_alias(AccountAlias=proposed_alias)
                except Exception as e:
                    log.error(e)


def scan_invited_accounts(log, org_client):
    """Return a list of handshake IDs"""
    response = org_client.list_handshakes_for_organization(
            Filter={'ActionType': 'INVITE'})
    handshakes = response['Handshakes']
    while 'NextToken' in response:
        response = org_client.list_handshakes_for_organization(
                Filter={'ActionType': 'INVITE'},
                NextToken=response['NextToken'])
        handshakes += response['Handshakes']
    log.debug(handshakes)
    return handshakes


def invite_account(log, args, org_client, deployed_accounts):
    """Invite account_id to join Org"""
    account_id = args['--invited-account-id']
    if not account_id:
        log.critical("option '--invited-account-id' not defined")
        sys.exit(1)
    if not valid_account_id(log, account_id):
        log.critical("option '--invited-account-id' must be a valid account Id")
        sys.exit(1)
    if lookup(deployed_accounts, 'Id', account_id):
        log.error("account %s already in organization" % account_id)
        return
    invited_accounts = scan_invited_accounts(log, org_client)
    account_invite = [invite for invite in invited_accounts 
            if lookup(invite['Parties'], 'Type', 'ACCOUNT', 'Id') == account_id]
    if account_invite:
        log.debug('account_invite: %s' % account_invite)
        invite_state = account_invite[0]['State']
        log.debug('invite_state: %s' % invite_state)
        if invite_state == 'ACCEPTED':
            log.error('Account %s has already accepted a previous invite' % account_id)
            return
        if invite_state in ['REQUESTED', 'OPEN']:
            log.error('Account %s has already been invited to Org and status is %s' % (
                    account_id, invite_state))
            return
    log.info("inviting account %s to join Org" % account_id)
    if args['--exec']:
        target = dict(Id=account_id , Type='ACCOUNT')
        handshake = org_client.invite_account_to_organization(Target=target)['Handshake']
        log.info('account invite handshake Id: %s' % handshake['Id'])
        return handshake
    return


def display_invited_accounts(log, org_client):
    invited_accounts = scan_invited_accounts(log, org_client)
    if invited_accounts:
        header = "Invited Accounts in Org:"
        overbar = '_' * len(header)
        log.info("\n%s\n%s\n" % (overbar, header))
        fmt_str = "{:16}{:12}{}"
        log.info(fmt_str.format('Id:', 'State:', 'Expires:'))
        for invite in invited_accounts:
            account_id = lookup(invite['Parties'], 'Type', 'ACCOUNT', 'Id')
            invite_state = invite['State']
            invite_expiration = invite['ExpirationTimestamp']
            log.info(fmt_str.format(account_id, invite_state, invite_expiration))


def display_provisioned_accounts(log, deployed_accounts, status):
    """
    Print report of currently deployed accounts in AWS Organization.
    status::    matches account status (ACTIVE|SUSPENDED)
    """
    if status not in ('ACTIVE', 'SUSPENDED'):
        raise RuntimeError("'status' must be one of ('ACTIVE', 'SUSPENDED')")
    sorted_account_names = sorted([a['Name'] for a in deployed_accounts
            if a['Status'] == status])
    if sorted_account_names:
        header = '%s Accounts in Org:' % status.capitalize()
        overbar = '_' * len(header)
        log.info("\n%s\n%s\n" % (overbar, header))
        fmt_str = "{:20}{:20}{:16}{}"
        log.info(fmt_str.format('Name:', 'Alias', 'Id:', 'Email:'))
        for name in sorted_account_names:
            account = lookup(deployed_accounts, 'Name', name)
            log.info(fmt_str.format(
                    name,
                    account['Alias'],
                    account['Id'],
                    account['Email']))


def unmanaged_accounts(log, deployed_accounts, account_spec):
    deployed_account_names = [a['Name'] for a in deployed_accounts] 
    spec_account_names = [a['Name'] for a in account_spec['accounts']]
    log.debug('deployed_account_names: %s' % deployed_account_names)
    log.debug('spec_account_names: %s' % spec_account_names)
    return [a for a in deployed_account_names if a not in spec_account_names]


def s3_object_for_accounts(s3_account_bucket, object_key, deployed_accounts):
    """
    Post the deployed_accounts list to s3 bucket
    """
    s3_client = boto3.client('s3')
    list_buckets_response = s3_client.list_buckets()
    bucket_list = [a['Name'] for a in list_buckets_response['Buckets']]
    if s3_account_bucket not in bucket_list:
        s3_client.create_bucket(
            ACL = 'private',
            Bucket = s3_account_bucket,
            CreateBucketConfiguration = {'LocationConstraint':'us-west-2'})
    s3_client.put_object(
	Bucket = s3_account_bucket,
	Key = object_key,
	Body = yamlfmt(deployed_accounts))
    return


def main():
    args = docopt(__doc__, version=awsorgs.__version__)
    log = get_logger(args)
    log.debug(args)
    args = load_config(log, args)
    credentials = get_assume_role_credentials(
            args['--master-account-id'],
            args['--org-access-role'])
    if isinstance(credentials, RuntimeError):
        log.critical(credentials)
        sys.exit(1)
    org_client = boto3.client('organizations', **credentials)
    root_id = get_root_id(org_client)
    deployed_accounts = scan_deployed_accounts(log, org_client)

    if args['report']:
        aliases = get_account_aliases(log, deployed_accounts, args['--org-access-role'])
        deployed_accounts = merge_aliases(log, deployed_accounts, aliases)
        display_provisioned_accounts(log, deployed_accounts, 'ACTIVE')
        display_provisioned_accounts(log, deployed_accounts, 'SUSPENDED')
        display_invited_accounts(log, org_client)
        s3_bucket = get_s3_bucket_name()
        s3_object_for_accounts(s3_bucket, S3_OBJECT_KEY, deployed_accounts)
        return

    account_spec = validate_spec(log, args)
    validate_master_id(org_client, account_spec)

    if args['create']:
        create_accounts(org_client, args, log, deployed_accounts, account_spec)
        unmanaged = unmanaged_accounts(log, deployed_accounts, account_spec)
        if unmanaged:
            log.warn("Unmanaged accounts in Org: %s" % (', '.join(unmanaged)))

    if args['alias']:
        queue_threads(log, deployed_accounts, set_account_alias,
                f_args=(log, args, account_spec, args['--org-access-role']),
                thread_count=10)

    if args['invite']:
        invite_account(log, args, org_client, deployed_accounts)
        

if __name__ == "__main__":
    main()



#!/usr/bin/env python


"""Manage recources in an AWS Organization.

Usage:
  awsorgs (report|organization) [--config FILE]
                                [--spec-dir PATH] 
                                [--master-account-id ID]
                                [--auth-account-id ID]
                                [--org-access-role ROLE]
                                [--exec] [-q] [-d|-dd]
  awsorgs (--help|--version)

Modes of operation:
  report         Display organization status report only.
  orgnanizaion   Run AWS Org management tasks per specification.

Options:
  -h, --help                Show this help message and exit.
  -V, --version             Display version info and exit.
  --config FILE             AWS Org config file in yaml format.
  --spec-dir PATH           Location of AWS Org specification file directory.
  --master-account-id ID    AWS account Id of the Org master account.    
  --auth-account-id ID      AWS account Id of the authentication account.
  --org-access-role ROLE    IAM role for traversing accounts in the Org.
  --exec                    Execute proposed changes to AWS Org.
  -q, --quiet               Repress log output.
  -d, --debug               Increase log level to 'DEBUG'.
  -dd                       Include botocore and boto3 logs in log stream.

"""


import yaml
import json
import time

import boto3
from docopt import docopt

import awsorgs
import awsorgs.utils
from awsorgs.utils import *
from awsorgs.spec import *


def validate_accounts_unique_in_org(log, root_spec):
    """
    Make sure accounts are unique across org
    """
    # recursively build mapping of accounts to ou_names
    def map_accounts(spec, account_map={}):
        if 'Accounts' in spec and spec['Accounts']:
            for account in spec['Accounts']:
                if account in account_map:
                    account_map[account].append(spec['Name'])
                else:
                    account_map[account] = [(spec['Name'])]
        if 'Child_OU' in spec and spec['Child_OU']:
            for child_spec in spec['Child_OU']:
                map_accounts(child_spec, account_map)
        return account_map
    # find accounts set to more than one OU
    unique = True
    for account, ou in list(map_accounts(root_spec).items()):
        if len(ou) > 1:
            log.error("Account '%s' set in multiple OU: %s" % (account, ou))
            unique = False
    if not unique:
        log.critical("Invalid org_spec: Do not assign accounts to multiple "
                "Organizatinal Units")
        sys.exit(1)


def enable_policy_type_in_root(org_client, root_id):
    """
    Ensure policy type 'SERVICE_CONTROL_POLICY' is enabled in the
    organization root.
    """
    p_type = org_client.list_roots()['Roots'][0]['PolicyTypes']
    if (not p_type or (p_type[0]['Type'] == 'SERVICE_CONTROL_POLICY'
            and p_type[0]['Status'] != 'ENABLED')):
        org_client.enable_policy_type(RootId=root_id, PolicyType='SERVICE_CONTROL_POLICY')


def get_parent_id(org_client, account_id):
    """
    Query deployed AWS organanization for 'account_id. Return the 'Id' of
    the parent OrganizationalUnit or 'None'.
    """
    parents = org_client.list_parents(ChildId=account_id)['Parents']
    try:
        len(parents) == 1
        return parents[0]['Id']
    except:
        raise RuntimeError("API Error: account '%s' has more than one parent: "
                % (account_id, parents))


def list_policies_in_ou (org_client, ou_id):
    """
    Query deployed AWS organanization.  Return a list (of type dict)
    of policies attached to OrganizationalUnit referenced by 'ou_id'.
    """
    policies_in_ou = org_client.list_policies_for_target(
            TargetId=ou_id, Filter='SERVICE_CONTROL_POLICY',)['Policies']
    return sorted([ou['Name'] for ou in policies_in_ou])


def scan_deployed_policies(org_client):
    """
    Return list of Service Control Policies deployed in Organization
    """
    return org_client.list_policies(Filter='SERVICE_CONTROL_POLICY')['Policies']


def scan_deployed_ou(log, org_client, root_id):
    """
    Recursively traverse deployed AWS Organization.  Return list of
    organizational unit dictionaries.  
    """
    def build_deployed_ou_table(org_client, parent_name, parent_id, deployed_ou):
        # recusive sub function to build the 'deployed_ou' table
        response = org_client.list_organizational_units_for_parent( ParentId=parent_id)
        child_ou = response['OrganizationalUnits']
        while 'NextToken' in response and response['NextToken']:
            response = org_client.list_organizational_units_for_parent(
                ParentId=parent_id, NextToken=response['NextToken'])
            child_ou += response['OrganizationalUnits']

        response = org_client.list_accounts_for_parent( ParentId=parent_id)
        accounts = response['Accounts']
        while 'NextToken' in response and response['NextToken']:
            response = org_client.list_accounts_for_parent(
                ParentId=parent_id, NextToken=response['NextToken'])
            accounts += response['Accounts']
        log.debug('parent_name: %s; ou: %s' % (parent_name, yamlfmt(child_ou)))
        log.debug('parent_name: %s; accounts: %s' % (parent_name, yamlfmt(accounts)))

        if not deployed_ou:
            deployed_ou.append(dict(
                    Name = parent_name,
                    Id = parent_id,
                    Child_OU = [ou['Name'] for ou in child_ou if 'Name' in ou],
                    Accounts = [acc['Name'] for acc in accounts if 'Name' in acc]))
        else:
            for ou in deployed_ou:
                if ou['Name'] == parent_name:
                    ou['Child_OU'] = [d['Name'] for d in child_ou]
                    ou['Accounts'] = [d['Name'] for d in accounts]
        for ou in child_ou:
            ou['ParentId'] = parent_id
            deployed_ou.append(ou)
            build_deployed_ou_table(org_client, ou['Name'], ou['Id'], deployed_ou)

    # build the table 
    deployed_ou = []
    build_deployed_ou_table(org_client, 'root', root_id, deployed_ou)
    log.debug(yamlfmt(deployed_ou))
    return deployed_ou


def display_provisioned_policies(org_client, log, deployed):
    """
    Print report of currently deployed Service Control Policies in
    AWS Organization.
    """
    header = "Provisioned Service Control Policies:"
    overbar = '_' * len(header)
    log.info("\n\n%s\n%s" % (overbar, header))
    for policy in deployed['policies']:
        log.info("\nName:\t\t%s" % policy['Name'])
        log.info("Description:\t%s" % policy['Description'])
        log.info("Id:\t%s" % policy['Id'])
        log.info("Content:")
        log.info(json.dumps(json.loads(org_client.describe_policy(
                PolicyId=policy['Id'])['Policy']['Content']),
                indent=2,
                separators=(',', ': ')))


def display_provisioned_ou(org_client, log, deployed_ou, parent_name, indent=0):
    """
    Recursive function to display the deployed AWS Organization structure.
    """
    # query aws for child orgs
    parent_id = lookup(deployed_ou, 'Name', parent_name, 'Id')
    child_ou_list = lookup(deployed_ou, 'Name', parent_name, 'Child_OU')
    child_accounts = lookup(deployed_ou, 'Name', parent_name, 'Accounts')
    # display parent ou name
    tab = '  '
    log.info(tab*indent + parent_name + ':')
    # look for policies
    policy_names = list_policies_in_ou(org_client, parent_id)
    if len(policy_names) > 0:
        log.info(tab*indent + tab + 'Policies: ' + ', '.join(policy_names))
    # look for accounts
    account_list = sorted(child_accounts)
    if len(account_list) > 0:
        log.info(tab*indent + tab + 'Accounts: ' + ', '.join(account_list))
    # look for child OUs
    if child_ou_list:
        log.info(tab*indent + tab + 'Child_OU:')
        indent+=2
        for ou_name in child_ou_list:
            # recurse
            display_provisioned_ou(org_client, log, deployed_ou, ou_name, indent)


def manage_account_moves(org_client, args, log, deployed, ou_spec, dest_parent_id):
    """
    Alter deployed AWS Organization.  Ensure accounts are contained
    by designated OrganizationalUnits based on OU specification.
    """
    if 'Accounts' in ou_spec and ou_spec['Accounts']:
        for account in ou_spec['Accounts']:
            account_id = lookup(deployed['accounts'], 'Name', account, 'Id')
            if not account_id:
                log.warn("Account '%s' not yet in Organization" % account)
            else:
                source_parent_id = get_parent_id(org_client, account_id)
                if dest_parent_id != source_parent_id:
                    log.info("Moving account '%s' to OU '%s'" %
                            (account, ou_spec['Name']))
                    if args['--exec']:
                        org_client.move_account(
                                AccountId=account_id,
                                SourceParentId=source_parent_id,
                                DestinationParentId=dest_parent_id)


def place_unmanged_accounts(org_client, args, log, deployed, account_list, dest_parent):
    """
    Move any unmanaged accounts into the default OU.
    """
    for account in account_list:
        account_id = lookup(deployed['accounts'], 'Name', account, 'Id')
        dest_parent_id   = lookup(deployed['ou'], 'Name', dest_parent, 'Id')
        source_parent_id = get_parent_id(org_client, account_id)
        if dest_parent_id and dest_parent_id != source_parent_id:
            log.info("Moving unmanged account '%s' to default OU '%s'" %
                    (account, dest_parent))
            if args['--exec']:
                org_client.move_account(
                        AccountId=account_id,
                        SourceParentId=source_parent_id,
                        DestinationParentId=dest_parent_id)


def manage_policies(org_client, args, log, deployed, org_spec):
    """
    Manage Service Control Policies in the AWS Organization.  Make updates
    according to the sc_policies specification.  Do not touch
    the default policy.  Do not delete an attached policy.
    """
    for p_spec in org_spec['sc_policies']:
        policy_name = p_spec['Name']
        log.debug("considering sc_policy: %s" % policy_name)
        # dont touch default policy
        if policy_name == org_spec['default_sc_policy']:
            continue
        policy = lookup(deployed['policies'], 'Name', policy_name)
        # delete existing sc_policy
        if ensure_absent(p_spec):
            if policy:
                log.info("Deleting policy '%s'" % (policy_name))
                # dont delete attached policy
                if org_client.list_targets_for_policy( PolicyId=policy_id)['Targets']:
                    log.error("Cannot delete policy '%s'. Still attached to OU" %
                            policy_name)
                elif args['--exec']:
                    org_client.delete_policy(PolicyId=policy['Id'])
            continue
        # create or update sc_policy
        statement = dict(Effect=p_spec['Effect'], Action=p_spec['Actions'], Resource='*')
        policy_doc = json.dumps(dict(Version='2012-10-17', Statement=[statement]))
        log.debug("spec sc_policy_doc: %s" % yamlfmt(policy_doc))
        # create new policy
        if not policy:
            log.info("Creating policy '%s'" % policy_name)
            if args['--exec']:
                org_client.create_policy(
                        Content=policy_doc,
                        Description=p_spec['Description'],
                        Name=p_spec['Name'],
                        Type='SERVICE_CONTROL_POLICY')
        # check for policy updates
        else:
            deployed_policy_doc = json.dumps(json.loads(org_client.describe_policy(
                    PolicyId=policy['Id'])['Policy']['Content']))
            log.debug("real sc_policy_doc: %s" % yamlfmt(deployed_policy_doc))
            if (p_spec['Description'] != policy['Description']
                or policy_doc != deployed_policy_doc):
                log.info("Updating policy '%s'" % policy_name)
                if args['--exec']:
                    org_client.update_policy(
                            PolicyId=policy['Id'],
                            Content=policy_doc,
                            Description=p_spec['Description'],)


def manage_policy_attachments(org_client, args, log, deployed, org_spec, ou_spec, ou_id):
    """
    Attach or detach specified Service Control Policy to a deployed 
    OrganizatinalUnit.  Do not detach the default policy ever.
    """
    # create lists policies_to_attach and policies_to_detach
    attached_policy_list = list_policies_in_ou(org_client, ou_id)
    if 'SC_Policies' in ou_spec and isinstance(ou_spec['SC_Policies'], list):
        spec_policy_list = ou_spec['SC_Policies']
    else:
        spec_policy_list = []
    policies_to_attach = [p for p in spec_policy_list
            if p not in attached_policy_list]
    policies_to_detach = [p for p in attached_policy_list
            if p not in spec_policy_list
            and p != org_spec['default_sc_policy']]
    # attach policies
    for policy_name in policies_to_attach:
        if not lookup(deployed['policies'],'Name',policy_name):
            if args['--exec']:
                raise RuntimeError("spec-file: ou_spec: policy '%s' not defined" %
                        policy_name)
        if not ensure_absent(ou_spec):
            log.info("Attaching policy '%s' to OU '%s'" % (policy_name, ou_spec['Name']))
            if args['--exec']:
                org_client.attach_policy(
                        PolicyId=lookup(deployed['policies'], 'Name', policy_name, 'Id'),
                        TargetId=ou_id)
    # detach policies
    for policy_name in policies_to_detach:
        log.info("Detaching policy '%s' from OU '%s'" % (policy_name, ou_spec['Name']))
        if args['--exec']:
            org_client.detach_policy(
                    PolicyId=lookup(deployed['policies'], 'Name', policy_name, 'Id'),
                    TargetId=ou_id)


def manage_ou(org_client, args, log, deployed, org_spec, ou_spec_list, parent_name):
    """
    Recursive function to manage OrganizationalUnits in the AWS
    Organization.
    """
    for ou_spec in ou_spec_list:
        # ou exists
        ou = lookup(deployed['ou'], 'Name', ou_spec['Name'])
        if ou:
            # check for child_ou. recurse before other tasks.
            if 'Child_OU' in ou_spec:
                manage_ou(org_client, args, log, deployed, org_spec,
                        ou_spec['Child_OU'], ou_spec['Name'])
            # check if ou 'absent'
            if ensure_absent(ou_spec):
                log.info("Deleting OU %s" % ou_spec['Name'])
                # error if ou contains anything
                error_flag = False
                for key in ['Accounts', 'SC_Policies', 'Child_OU']:
                    if key in ou and ou[key]:
                        log.error("Can not delete OU '%s'. deployed '%s' exists." %
                                (ou_spec['Name'], key))
                        error_flag = True
                if error_flag:
                    continue
                elif args['--exec']:
                    org_client.delete_organizational_unit(
                            OrganizationalUnitId=ou['Id'])
            # manage account and sc_policy placement in OU
            else:
                manage_policy_attachments(org_client, args, log,
                        deployed, org_spec, ou_spec, ou['Id'])
                manage_account_moves(org_client, args, log, deployed, ou_spec, ou['Id'])
        # create new OU
        elif not ensure_absent(ou_spec):
            log.info("Creating new OU '%s' under parent '%s'" %
                    (ou_spec['Name'], parent_name))
            if args['--exec']:
                new_ou = org_client.create_organizational_unit(
                        ParentId=lookup(deployed['ou'],'Name',parent_name,'Id'),
                        Name=ou_spec['Name'])['OrganizationalUnit']
                # account and sc_policy placement
                manage_policy_attachments(org_client, args, log,
                        deployed, org_spec, ou_spec, new_ou['Id'])
                manage_account_moves(org_client, args, log, deployed, ou_spec, new_ou['Id'])
                # recurse if child OU
                if ('Child_OU' in ou_spec and isinstance(new_ou, dict)
                        and 'Id' in new_ou):
                    manage_ou(org_client, args, log, deployed, org_spec,
                            ou_spec['Child_OU'], new_ou['Name'])


def main():
    args = docopt(__doc__, version=awsorgs.__version__)
    log = get_logger(args)
    log.debug(args)
    args = load_config(log, args)
    credentials = get_assume_role_credentials(
            args['--master-account-id'],
            args['--org-access-role'])
    if isinstance(credentials, RuntimeError):
        log.critical(credentials)
        sys.exit(1)
    org_client = boto3.client('organizations', **credentials)
    root_id = get_root_id(org_client)
    deployed = dict(
            policies = scan_deployed_policies(org_client),
            accounts = scan_deployed_accounts(log, org_client),
            ou = scan_deployed_ou(log, org_client, root_id))

    if args['report']:
        header = 'Provisioned Organizational Units in Org:'
        overbar = '_' * len(header)
        log.info("\n%s\n%s" % (overbar, header))
        display_provisioned_ou(org_client, log, deployed['ou'], 'root')
        display_provisioned_policies(org_client, log, deployed)

    if args['organization']:
        org_spec = validate_spec(log, args)
        root_spec = lookup(org_spec['organizational_units'], 'Name', 'root')
        validate_master_id(org_client, org_spec)
        validate_accounts_unique_in_org(log, root_spec)
        managed = dict(
                accounts = search_spec(root_spec, 'Accounts', 'Child_OU'),
                ou = search_spec(root_spec, 'Name', 'Child_OU'),
                policies = [p['Name'] for p in org_spec['sc_policies']])

        # ensure default_sc_policy is considered 'managed'
        if org_spec['default_sc_policy'] not in managed['policies']:
            managed['policies'].append(org_spec['default_sc_policy'])
        enable_policy_type_in_root(org_client, root_id)
        manage_policies(org_client, args, log, deployed, org_spec)

        # rescan deployed policies
        deployed['policies'] = scan_deployed_policies(org_client)
        manage_ou(org_client, args, log, deployed, org_spec,
                org_spec['organizational_units'], 'root')

        # check for unmanaged resources
        for key in list(managed.keys()):
            unmanaged= [a['Name'] for a in deployed[key] if a['Name'] not in managed[key]]
            if unmanaged:
                log.warn("Unmanaged %s in Organization: %s" % (key,', '.join(unmanaged)))
                if key ==  'accounts':
                    # append unmanaged accounts to default_ou
                    place_unmanged_accounts(org_client, args, log, deployed,
                            unmanaged, org_spec['default_ou'])


if __name__ == "__main__":
    main()



"""aws-orgs setup"""

from awsorgs import __version__
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aws-orgs',
    version=__version__,
    description='Tools to manage AWS Organizations',
    long_description=long_description,
    url='https://github.com/ashleygould/aws-orgs',
    author='Ashley Gould',
    author_email='agould@ucop.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='aws organizations',
    packages=find_packages(exclude=['scratch', 'notes']),
    install_requires=[
        'boto3', 
        'docopt', 
        'PyYAML', 
        'passwordgenerator',
        'cerberus',
    ],
    package_data={
        'awsorgs': [
            'samples/*.yaml',
            'data/*',
        ],
    },
    entry_points={
        'console_scripts': [
            'awsorgs=awsorgs.orgs:main',
            'awsaccounts=awsorgs.accounts:main',
            'awsauth=awsorgs.auth:main',
            'awsloginprofile=awsorgs.loginprofile:main',
            'awsorgs-accessrole=awsorgs.accessrole:main',
        ],
    },

)



#!/usr/bin/env python
import boto3


def get_passwd_policy():
    client = boto3.client('iam')
    try:
        return client.get_account_password_policy()['PasswordPolicy']
    except client.exceptions.NoSuchEntityException:
        return "password policy not implemented"

def set_passwd_policy():
    client = boto3.client('iam')
    client.update_account_password_policy(
        MinimumPasswordLength=8,
        RequireSymbols=True,
        RequireNumbers=True,
        RequireUppercaseCharacters=True,
        RequireLowercaseCharacters=True,
        AllowUsersToChangePassword=True,
        MaxPasswordAge=180,
        PasswordReusePrevention=6,
        #HardExpiry=True|False
    )
    return

def delete_passwd_policy():
    client = boto3.client('iam')
    client.delete_account_password_policy()
    return 



if __name__ == '__main__':
    #response = set_passwd_policy()
    #response = get_passwd_policy()
    #print(yamlfmt(response))
    #response = delete_passwd_policy()
    #response = get_passwd_policy()
    #print(yamlfmt(response))
